@@ -1,4 +1,4 @@
-(* $Id: netcgi.ml,v 1.10 2002/10/24 23:47:48 stolpmann Exp $
+(* $Id: netcgi.ml,v 1.12 2004/07/01 12:56:20 stolpmann Exp $
  * ----------------------------------------------------------------------
  *
  *)
@@ -7,6 +7,49 @@ open Netcgi_env ;;
 open Netcgi_types ;;
 open Netchannels ;;
 
+
+let status_line (s:status) =
+  match s with
+      `Ok -> "200 OK"
+    | `Created -> "201 Created"
+    | `Accepted -> "202 Accepted"
+    | `Non_authoritative -> "203 Non-Authoritative Information"
+    | `No_content -> "204 No content"
+    | `Reset_content -> "205 Reset Content"
+    | `Partial_content -> "206 Partial Content"
+    | `Multiple_choices -> "300 Multiple Choices"
+    | `Moved_permanently -> "301 Moved Permanently"
+    | `Found -> "302 Found"
+    | `See_other -> "303 See Other"
+    | `Not_modified -> "304 Not Modified"
+    | `Use_proxy -> "305 Use Proxy"
+    | `Temporary_redirect -> "307 Temporary Redirect"
+    | `Bad_request -> "400 Bad Request"
+    | `Unauthorized -> "401 Unauthorized"
+    | `Payment_required -> "402 Payment Required"
+    | `Forbidden -> "403 Forbidden"
+    | `Not_found -> "404 Not Found"
+    | `Method_not_allowed -> "405 Method Not Allowed"
+    | `Not_acceptable -> "406 Not Acceptable"
+    | `Proxy_auth_required -> "407 Proxy Authentication Required"
+    | `Request_timeout -> "408 Request Timeout"
+    | `Conflict -> "409 Conflict"
+    | `Gone -> "410 Gone"
+    | `Length_required -> "411 Length Required"
+    | `Precondition_failed -> "412 Precondition Failed"
+    | `Request_entity_too_large -> "413 Request Entity Too Large"
+    | `Request_uri_too_long -> "414 Request-URI Too Long"
+    | `Unsupported_media_type -> "415 Unsupported Media Type"
+    | `Requested_range_not_satisfiable -> "416 Requested Range Not Satisfiable"
+    | `Expectation_failed -> "417 Expectation Failed"
+    | `Internal_server_error -> "500 Internal Server Error"
+    | `Not_implemented -> "501 Not Implemented"
+    | `Bad_gateway -> "502 Bad Gateway"
+    | `Service_unavailable -> "503 Service Unavailable"
+    | `Gateway_timeout -> "504 Gateway Timeout"
+    | `Http_version_not_supported -> "505 HTTP Version Not Supported"
+;;
+
 class simple_argument ?(ro=false) name value : cgi_argument =
 object(self)
   inherit Netmime.memory_mime_body ~ro value
@@ -108,7 +151,6 @@ object
   method flush =          ch # flush
   method close_out =      ch # close_out
   method pos_out =        ch # pos_out
-  method _rep_out =       ch # _rep_out
 
   method commit_work ()  = ch # flush(); f_after()
   method rollback_work() = ch # output_string sep
@@ -135,7 +177,6 @@ object
   method flush =          ch # flush
   method close_out =      ch # close_out
   method pos_out =        ch # pos_out
-  method _rep_out =       ch # _rep_out
   method rollback_work =  ch # rollback_work
 
   method commit_work () = f_before(); ch # commit_work(); f_after()
@@ -396,7 +437,7 @@ object (self)
     (match status with
 	 None -> ()
        | Some s -> 
-	   environ # set_output_header_field "Status" (Cgi.status_line s)
+	   environ # set_output_header_field "Status" (status_line s)
     );
     (match cache with
 	 `Unspecified -> ()
@@ -706,6 +747,16 @@ let tempfile_transactional_optype =
  * History:
  * 
  * $Log: netcgi.ml,v $
+ * Revision 1.12  2004/07/01 12:56:20  stolpmann
+ * 	Removed dependency on Cgi.
+ *
+ * Revision 1.11  2004/05/30 21:27:14  stolpmann
+ * 	Supporting the standard for class-based I/O. This also means
+ * the semantics of [input] and [output] methods have changed (return
+ * value 0, End_of_file).
+ * 	Removed _rep_in, _rep_out.
+ * 	Removed Netstream.Deprecated.
+ *
  * Revision 1.10  2002/10/24 23:47:48  stolpmann
  * 	Support for the HEAD method.
  * 	Workaround for a bug in MSIE: Empty cookies are represented
