@@ -1,4 +1,4 @@
-(* $Id: mimestring.ml,v 2.12 2002/02/02 16:10:47 stolpmann Exp $
+(* $Id: mimestring.ml,v 2.15 2004/07/06 15:20:53 stolpmann Exp $
  * ----------------------------------------------------------------------
  *
  *)
@@ -517,18 +517,22 @@ let scan_next_token ((spec,target) as scn) =
     let k = ref 0 in
     let line = ref target.scanner_line in
     let linepos = ref target.scanner_linepos in
+    let esc = ref false in
     for i = i0 to i1 do
       let c = s.[i] in
       match c with
-	  '\\' when i < i1 &&  not spec.opt_no_backslash_escaping -> ()
+	  '\\' when i < i1 &&  not spec.opt_no_backslash_escaping && not !esc ->
+	    esc := true
 	| '\n' ->
 	    line := !line + 1;
 	    linepos := i+1;
 	    r.[ !k ] <- c; 
-	    incr k
+	    incr k;
+	    esc := false
 	| _ -> 
 	    r.[ !k ] <- c; 
-	    incr k
+	    incr k;
+	    esc := false
     done;
     assert (!k = n);
     r, !line, !linepos
@@ -726,6 +730,24 @@ let scan_encoded_text_value s =
 ;;
 
 
+let split_mime_type ct_type =
+  let specials = specials_rfc2045 @ [ '"'; '['; ']' ] in
+  let scn = create_mime_scanner specials [] ct_type in
+  let rec collect () =
+    match scan_token scn with
+	_, End ->
+	  []
+      | _, tok ->
+	  tok :: collect()
+  in
+  match collect() with
+      [ Atom main_type; Special '/'; Atom sub_type ] ->
+	(String.lowercase main_type, String.lowercase sub_type)
+    | _ ->
+	failwith "Mimestring.split_mime_type"
+;;
+
+
 type s_param =
     P_encoded of (string * string * string)
   | P of string
@@ -961,6 +983,166 @@ let scan_mime_type_ep s options =
 ;;
 
 
+let generate_encoded_words cs lang enc_tok text len1 len =
+  (* Generate a list of encoded words for the charset [cs], the
+   * language [lang], encoded as [enc_tok]. The text is passed
+   * in [text] (encoded only in [cs], [enc_tok] is not applied). 
+   *
+   * [len1] are the number of bytes of the first token,
+   * [len] are the number of bytes for the following tokens.
+   * If [len1] is too small for at least one token, [len]
+   * is also used for the first token.
+   *
+   * If [cs] is not known to [Netconversion], it is assumed
+   * it is a single-byte encoding.
+   *)
+  (* Determine prefix and suffix of encoded words: *)
+  let prefix = 
+    "=?" ^ cs ^ 
+    (if lang <> "" then "*" ^ lang else "") ^
+    ( match enc_tok with
+	  "Q"|"q" -> "?Q?"
+	| "B"|"b" -> "?B?"
+	| _ -> failwith "Mimestring.write_value: Unknown encoding"
+    ) in
+  let suffix = "?=" in
+  let prefix_len = String.length prefix in
+  let suffix_len = String.length suffix in
+
+  let encode s =
+    match enc_tok with
+	"Q"|"q" -> Netencoding.Q.encode s
+      | "B"|"b" -> Netencoding.Base64.encode s
+      | _ -> assert false (* already caught above *)
+  in
+
+  let find_tok_single_byte k l =
+    (* k: The position where to determine the token
+     * l: Desired length of the token
+     *
+     * Returns the position of the end of the token
+     *)
+    let p = ref k in
+    let l_cur() = 
+      let slice = String.sub text k (!p - k) in
+      prefix_len + String.length (encode slice) + suffix_len
+    in
+    if l_cur() > l then
+      k
+    else (
+      try
+	while l_cur() <= l do
+	  incr p;
+	  if !p >= String.length text then raise Exit
+	done;
+	decr p;
+	!p
+      with
+	  Exit ->
+	    !p  (* at the end of the string *)
+    )
+  in
+
+  let rec generate_single_byte k l =
+    (* k: Position in enc_text
+     * l: Desired length of the token
+     *)
+    if k >= String.length text then
+      []
+    else (
+      let j = ref(find_tok_single_byte k l) in
+      if k = !j then (
+	(* l is too short, use len instead *)
+	j := find_tok_single_byte k len;
+	(* Still too short: Process only one char *)
+	if k = !j then (
+	  j := k+1
+	)
+      );
+      let tok =
+	prefix ^ 
+	encode(String.sub text k (!j - k)) ^
+	suffix in
+      tok :: generate_single_byte !j len
+    )
+  in
+
+  let find_tok_multi_byte cursor l =
+    (* cursor: The cursor moving around text
+     * l: Desired length of the token
+     *
+     * Moves the cursor to the end of the token
+     *)
+    let k_start = Netconversion.cursor_pos cursor in
+    let l_cur() = 
+      let k_cur = Netconversion.cursor_pos cursor in
+      let slice = String.sub text k_start (k_cur - k_start) in
+      prefix_len + String.length (encode slice) + suffix_len
+    in
+    if l_cur() > l then
+      ()
+    else (
+      try
+	while l_cur() <= l do
+	  Netconversion.move cursor
+	done;
+	Netconversion.move ~num:(-1) cursor
+      with
+	  Netconversion.Cursor_out_of_range ->
+	    ()  (* at the end of the string *)
+    )
+  in
+
+  let rec generate_multi_byte cursor l =
+    (* cursor: The cursor moving around text
+     * l: Desired length of the token
+     *)
+    if Netconversion.cursor_at_end cursor then
+      []
+    else (
+      let k_start = Netconversion.cursor_pos cursor in
+      find_tok_multi_byte cursor l;
+      if k_start = Netconversion.cursor_pos cursor then (
+	(* l is too short, use len instead *)
+	find_tok_multi_byte cursor len;
+	(* Still too short: Process only one char *)
+	if k_start = Netconversion.cursor_pos cursor then (
+	  Netconversion.move cursor
+	)
+      );
+      let k_end = Netconversion.cursor_pos cursor in
+      let tok =
+	prefix ^ 
+	encode(String.sub text k_start (k_end - k_start)) ^
+	suffix in
+      tok :: generate_multi_byte cursor len
+    )
+  in
+
+  (* Check if the charset(encoding) is known, and if so, is available. *)
+  let is_single_byte, e =
+    try
+      let enc = Netconversion.encoding_of_string cs in
+                (* or Failure *)
+      ( Netconversion.is_single_byte enc ||
+	not (List.mem enc (Netconversion.available_input_encodings())),
+	enc )
+    with
+	Failure _ -> (true, `Enc_empty) in
+
+  (* For single-byte, or when the multi-byte encoding is not
+   * available, use [generate_single_byte]. 
+   * For available multi-byte encodings, use [generate_multi_byte]
+   * to split the string only at allowed positions.
+   *)
+  if is_single_byte then
+    generate_single_byte 0 len1
+  else
+    let cursor = Netconversion.create_cursor e text in
+    generate_multi_byte cursor len1
+;;
+
+
 exception Line_too_long;;
 
 let write_value ?maxlen1 ?maxlen ?hardmaxlen1 ?hardmaxlen 
@@ -1115,80 +1297,24 @@ let write_value ?maxlen1 ?maxlen ?hardmaxlen1 ?hardmaxlen
 	  ([], tl)
   in
   
-  let output_encword cs lang enc_tok enc_text pos len =
-    output_nobreak "=?";
-    output_nobreak cs;
-    if lang <> "" then begin
-      output_nobreak "*";
-      output_nobreak lang;
-    end;
-    ( match enc_tok with
-	  `Q -> output_nobreak "?Q?"
-	| `B -> output_nobreak "?B?"
-    );
-    output_nobreak ~pos ~len enc_text;
-    output_nobreak "?=";
-  in
-
-  let minlen_encword cs lang enc_tok =
-    String.length cs + 
-    (if lang <> "" then String.length lang + 1 else 0) +
-    7
-  in
-
-  let rec output_encoded cs lang enc_tok enc_text pos =
-    assert(do_folding);
-    let remaining = String.length enc_text - pos in
-    if remaining > 0 then begin
-      (* pos>0: This is not the first invocation of the recursion. Print
-       * a break to separate from the previous encoded word
-       *)
-      if pos > 0 then begin
-	output_break();
-	output_nobreak " ";
-      end;
-      (* How much space do we have? *)
-      let space = 
-	min (max (!still_unused - Buffer.length break_buffer - 
-		  minlen_encword cs lang enc_tok) 0) remaining in
-      (* If space<4, it is better to calculate with maxlen: *)
-      let space' =
-	if space < 4 then
-	  let l = get_maxlen() in
-	  min (max (l - Buffer.length break_buffer -
-		    minlen_encword cs lang enc_tok) 0) remaining 
-	else
-	  space
-      in
-      (* We could print [space'] bytes of [enc_text] here. *)
-      let n =
-	match enc_tok with
-	    `B -> 
-	      (* round to the next number that is divisible by 4 *)
-	      space' land (-4)
-	  | `Q ->
-	      (* round to the next encoding border *)
-	      if space' >= 4 then begin
-		if enc_text.[pos + space' - 1] = '=' then space' - 1 else
-		  if enc_text.[pos + space' - 2] = '=' then space' - 2 else
-		    space'
-	      end
-	      else space'
-      in
-      (* Printing n bytes would be ok, but we have to avoid n=0 *)
-      let n' =
-	if n=0 then begin
-	  match enc_tok with
-	      `B -> 4
-	    | `Q -> if enc_text.[pos] = '=' then 3 else 1
-	end
-	else n
-      in
-      assert(remaining >= n');
-      (* Finally print: *)
-      output_encword cs lang enc_tok enc_text pos n';
-      output_encoded cs lang enc_tok enc_text (pos+n');
-    end
+  let output_encoded cs lang enc_tok text =
+    let len1 =
+      max 0 (!still_unused - Buffer.length break_buffer) in
+    let len =
+      if do_folding then get_maxlen () - 1 else max_int in
+    let words = 
+      generate_encoded_words cs lang enc_tok text len1 len in
+    let first = ref true in
+    List.iter
+      (fun word ->
+	 if not !first then (
+	   output_break();
+	   output_nobreak " ";
+	 );
+	 output_nobreak word;
+	 first := false
+      )
+      words
   in    
 
   let rec output_tokens last tl =
@@ -1240,16 +1366,7 @@ let write_value ?maxlen1 ?maxlen ?hardmaxlen1 ?hardmaxlen
 	  separate_words last;
 	  let (textlist, tl'') = collect_words cs lang enc tl in
 	  let text = String.concat "" textlist in
-	  let enc_text, enc_tok =
-	    match enc with
-		"Q"|"q" -> Netencoding.Q.encode text, `Q
-	      | "B"|"b" -> Netencoding.Base64.encode text, `B
-	      | _ -> failwith "Mimestring.write_value: Unknown encoding"
-	  in
-	  if do_folding then
-	    output_encoded cs lang enc_tok enc_text 0
-	  else
-	    output_encword cs lang enc_tok enc_text 0 (String.length enc_text);
+	  output_encoded cs lang enc text;
 	  output_tokens cur tl''
       | End :: tl' ->
 	  output_tokens End tl'
@@ -1394,6 +1511,18 @@ let rec param_tokens ?(maxlen=max_int) pl =
 ;;
 
 
+let split_uri uri =
+  let rec split k =
+    if k >= String.length uri then
+      []
+    else
+      let k' = min (k + 40) (String.length uri) in
+      (QString (String.sub uri k (k' - k)) :: split k')
+  in
+  split 0
+;;
+
+
 let lf_re = S.regexp "[\n]";;
 
 
@@ -1620,6 +1749,15 @@ let create_boundary ?(random = []) ?(nr = 0) () =
  * History:
  *
  * $Log: mimestring.ml,v $
+ * Revision 2.15  2004/07/06 15:20:53  stolpmann
+ * 	Bugfixes: write_value now correctly folds long encoded words.
+ *
+ * Revision 2.14  2004/07/04 18:51:54  stolpmann
+ * 	Added: split_mime_type, split_uri
+ *
+ * Revision 2.13  2004/05/30 21:22:00  stolpmann
+ * 	Fixed bug when parsing backslashes within double quotes
+ *
  * Revision 2.12  2002/02/02 16:10:47  stolpmann
  * 	Bugfix: read_header recognizes empty headers.
  *
