@@ -1,4 +1,4 @@
-(* $Id: netchannels.ml,v 1.15 2002/11/01 17:11:07 stolpmann Exp $
+(* $Id: netchannels.ml,v 1.18 2004/07/11 15:40:29 stolpmann Exp $
  * ----------------------------------------------------------------------
  *
  *)
@@ -7,23 +7,37 @@ exception Closed_channel
 exception Buffer_underrun
 exception Command_failure of Unix.process_status
 
-class type raw_in_channel = object
+class type rec_in_channel = object
   method input : string -> int -> int -> int
   method close_in : unit -> unit
+end
+
+class type raw_in_channel = object
+  inherit rec_in_channel
   method pos_in : int             (* number of read characters *)
 end
 
+type input_result =
+    [ `Data of int
+    | `Separator of string
+    ]
+
 class type enhanced_raw_in_channel =
 object 
   inherit raw_in_channel
   method private enhanced_input_line : unit -> string
+  method private enhanced_input : string -> int -> int -> input_result
+end
+
+class type rec_out_channel = object
+  method output : string -> int -> int -> int
+  method close_out : unit -> unit
+  method flush : unit -> unit
 end
 
 class type raw_out_channel = object
-  method output : string -> int -> int -> int
-  method close_out : unit -> unit
+  inherit rec_out_channel
   method pos_out : int             (* number of written characters *)
-  method flush : unit -> unit
 end
 
 class type raw_io_channel = object
@@ -37,10 +51,6 @@ class type compl_in_channel = object
   method input_char : unit -> char
   method input_line : unit -> string
   method input_byte : unit -> int
-
-  (* Transitional method: *)
-  method _rep_in : [ `Chan of in_channel | `Other ]
-     (* THIS METHOD IS DEPRECATED AND WILL BE REMOVED SOON *)
 end
 
 class type in_obj_channel = object
@@ -58,10 +68,6 @@ class type compl_out_channel = object
   method output_buffer : Buffer.t -> unit
   method output_channel : ?len:int -> in_obj_channel -> unit
       (* ~len: optionally limit the number of bytes *)
-
-  (* Transitional method: *)
-  method _rep_out : [ `Chan of out_channel | `Other ]
-     (* THIS METHOD IS DEPRECATED AND WILL BE REMOVED SOON *)
 end
 
 
@@ -102,6 +108,67 @@ type extended_error_behavior =
 type close_mode = [ `Commit | `Rollback ];;
 
 
+(* Delegation *)
+
+class rec_in_channel_delegation ?(close=true) (ch:rec_in_channel) =
+object(self)
+  method input = ch#input
+  method close_in() = if close then ch#close_in()
+end
+
+
+class raw_in_channel_delegation ?(close=true) (ch:raw_in_channel) =
+object(self)
+  method input = ch#input
+  method close_in() = if close then ch#close_in()
+  method pos_in = ch#pos_in
+end
+
+
+class in_obj_channel_delegation ?(close=true) (ch:in_obj_channel) =
+object(self)
+  method input = ch#input
+  method close_in() = if close then ch#close_in()
+  method pos_in = ch#pos_in
+  method really_input = ch#really_input
+  method input_char = ch#input_char
+  method input_line = ch#input_line
+  method input_byte = ch#input_byte
+end
+
+
+class rec_out_channel_delegation ?(close=true) (ch:rec_out_channel) =
+object(self)
+  method output = ch#output
+  method close_out() = if close then ch#close_out()
+  method flush = ch#flush
+end
+
+
+class raw_out_channel_delegation ?(close=true) (ch:raw_out_channel) =
+object(self)
+  method output = ch#output
+  method close_out() = if close then ch#close_out()
+  method flush = ch#flush
+  method pos_out = ch#pos_out
+end
+
+
+class out_obj_channel_delegation ?(close=true) (ch:out_obj_channel) =
+object(self)
+  method output = ch#output
+  method close_out() = if close then ch#close_out()
+  method flush = ch#flush
+  method pos_out = ch#pos_out
+  method really_output = ch#really_output
+  method output_char = ch#output_char
+  method output_string = ch#output_string
+  method output_byte = ch#output_byte
+  method output_buffer = ch#output_buffer
+  method output_channel = ch#output_channel
+end
+
+
 (****************************** input ******************************)
 
 class input_channel ch (* : in_obj_channel *) =
@@ -114,7 +181,12 @@ object (self)
 
   method input buf pos len = 
     if closed then self # complain_closed();
-    Pervasives.input ch buf pos len 
+    try
+      if len=0 then raise Sys_blocked_io;
+      let n = Pervasives.input ch buf pos len in
+      if n=0 then raise End_of_file else n
+    with
+	Sys_blocked_io -> 0
 
   method really_input buf pos len = 
     if closed then self # complain_closed();
@@ -139,8 +211,6 @@ object (self)
   method pos_in =
     if closed then self # complain_closed();
     Pervasives.pos_in ch 
-
-  method _rep_in = (`Chan ch : [ `Chan of in_channel | `Other ])
 end
 ;;
 
@@ -191,7 +261,7 @@ object (self)
     
     str_pos <- str_pos + n;
 
-    n
+    if n=0 && len>0 then raise End_of_file else n
 
 
   method really_input buf pos len =
@@ -243,8 +313,6 @@ object (self)
     if closed then self # complain_closed();
     str_pos
 
-
-  method _rep_in = `Other
 end
 ;;
 
@@ -272,8 +340,8 @@ object (self)
       invalid_arg "input";
 
     let n = min len (Netbuffer.length b) in
-    if n = 0 then begin
-      if eof then 0 else raise Buffer_underrun
+    if n = 0 && len>0 then begin
+      if eof then raise End_of_file else raise Buffer_underrun
     end
     else begin
       Netbuffer.blit b 0 buf pos n;
@@ -297,7 +365,6 @@ object (self)
     if closed then self # complain_closed();
     let s = String.create 1 in
     match self # input s 0 1 with
-	0 -> raise End_of_file
       | 1 -> s.[0]
       | _ -> assert false
 
@@ -338,9 +405,6 @@ object (self)
     ch_pos
 
 
-  method _rep_in = `Other
-
-
   method shutdown() = eof <- true
 end
 ;;
@@ -354,7 +418,12 @@ let create_input_netbuffer b =
 
 let lexbuf_of_in_obj_channel (objch : in_obj_channel) : Lexing.lexbuf =
   let fill_buffer buf len =
-    objch # input buf 0 len
+    try
+      let n = objch # input buf 0 len in
+      if n=0 then failwith "Netchannels.lexbuf_of_in_obj_channel: No data (non-blocking I/O?)";
+      n
+    with
+	End_of_file -> 0
   in
   Lexing.from_function fill_buffer
 ;;
@@ -372,7 +441,8 @@ let string_of_in_obj_channel (objch : in_obj_channel) : string =
     while true do
       let s = String.create slen in
       let n = objch # input s 0 slen in
-      if n = 0 then raise End_of_file;
+      if n = 0 then 
+	failwith "Netchannels.string_of_in_obj_channel: No data (non-blocking I/O?)";
       k := !k + n;
       if n < slen then
 	l := (String.sub s 0 n) :: !l
@@ -419,7 +489,7 @@ object (self)
     let rec read_rest n =
       if n < len then
 	let m = self # input s (pos+n) (len-n) in
-	if m = 0 then raise End_of_file;
+	if m = 0 then raise Sys_blocked_io;
 	read_rest (n+m)
       else
 	()
@@ -440,19 +510,299 @@ object (self)
     let s = String.create 1 in
     let b = Buffer.create 80 in
     let m = self # input s 0 1 in
-    if m = 0 then raise End_of_file;
+    if m = 0 then raise Sys_blocked_io;
     while s.[0] <> '\n' do
       Buffer.add_char b s.[0];
-      let m = self # input s 0 1 in
-      if m = 0 then s.[0] <- '\n'
+      try
+	let m = self # input s 0 1 in
+	if m = 0 then raise Sys_blocked_io;
+      with
+	  End_of_file ->
+	    s.[0] <- '\n'
     done;
     Buffer.contents b
 
-  method _rep_in = (`Other : [ `Chan of in_channel | `Other ])
 end
 ;;
 
 
+class lift_raw_in_channel r =
+object(self)
+  inherit augment_raw_in_channel
+
+  method input s p l =
+    r # input s p l
+
+  method close_in () =
+    r # close_in()
+
+  method pos_in =
+    r # pos_in
+
+end;;
+
+
+class lift_rec_in_channel ?(start_pos_in = 0) (r : rec_in_channel) =
+object(self)
+  inherit augment_raw_in_channel
+
+  val mutable closed = false
+  val mutable pos_in = start_pos_in
+
+  method input s p l =
+    if closed then raise Closed_channel;
+    let n = r # input s p l in
+    pos_in <- pos_in + n;
+    n
+
+  method close_in () =
+    if closed then raise Closed_channel;
+    closed <- true;
+    r # close_in()
+
+  method pos_in =
+    if closed then raise Closed_channel;
+    pos_in
+
+end;;
+
+
+type eol_status =
+    EOL_not_found
+  | EOL_partially_found of int (* Position *)
+  | EOL_found of int * int     (* Position, length *)
+
+
+class buffered_raw_in_channel
+      ?(eol = [ "\n" ])
+      ?(buffer_size = 4096)
+      (ch : raw_in_channel) : enhanced_raw_in_channel =
+object (self)
+  val out = ch
+  val bufsize = buffer_size
+  val buf = String.create buffer_size
+  val mutable bufpos = 0
+  val mutable buflen = 0
+  val mutable eof = false
+  val mutable closed = false
+
+  initializer
+    if List.exists(fun s -> s = "") eol then
+      invalid_arg "Netchannels.buffered_raw_in_channel";
+    if List.exists(fun s -> String.length s > buffer_size) eol then
+      invalid_arg "Netchannels.buffered_raw_in_channel";
+
+  method input s pos len =
+    if closed then raise Closed_channel;
+    if len > 0 then (
+      if bufpos = buflen then (
+	self # refill();
+      );
+      let n = min len (buflen - bufpos) in
+      String.blit buf bufpos s pos n;
+      bufpos <- bufpos + n;
+      n
+    )
+    else 0
+
+  method private refill() =
+    let d = bufpos in
+    if d > 0 && d < buflen then (
+      String.blit buf d buf 0 (buflen-d)
+    );
+    bufpos <- 0;
+    buflen <- buflen - d;
+    try
+      assert(bufsize > buflen);  (* otherwise problems... *)
+      let n = ch # input buf buflen (bufsize-buflen) in  (* or End_of_file *)
+      if n = 0 then raise Sys_blocked_io;
+      buflen <- buflen+n;
+    with
+	End_of_file as exn ->
+	  eof <- true;
+	  raise exn
+
+
+  method close_in () =
+    ch # close_in();
+    closed <- true
+
+  method pos_in =
+    (ch # pos_in) - (buflen - bufpos)
+
+
+  method private find_eol() =
+    (* Try all strings from [eol] in turn. For every string we may
+     * have three results:
+     * - Not found
+     * - Partially found
+     * - Found
+     * The eol delimiter is only found if there are no partial
+     * results, and at least one positive result. The longest
+     * string is taken.
+     *)
+
+    let find_this_eol eol =
+      (* Try to find the eol string [eol] in [buf] starting at
+       * [bufpos] up to [buflen]. Return [eol_status].
+       *)
+      let eol0 = eol.[0] in
+      try
+	let k = String.index_from buf bufpos eol0 in (* or Not_found *)
+	if k>=buflen then raise Not_found;
+	let k' = min buflen (k+String.length eol) in
+	let s = String.sub buf k (k' - k) in
+	if s = eol then
+	  EOL_found(k, String.length eol)
+	else
+	  if not eof && String.sub eol 0 (String.length s) = s then
+	    EOL_partially_found k
+	  else
+	    EOL_not_found
+      with
+	    Not_found -> EOL_not_found
+    in  
+
+    let rec find_best_eol best eol_result =
+      match eol_result with
+	  EOL_not_found :: eol_result' ->
+	    find_best_eol best eol_result'
+	| EOL_partially_found pos as r :: eol_result' ->
+	    ( match best with
+		  EOL_partially_found pos' ->
+		    if pos < pos' then
+		      find_best_eol r eol_result'
+		    else
+		      find_best_eol best eol_result'
+		| _ ->
+		    find_best_eol r eol_result'
+	    )
+	| EOL_found(pos,len) as r :: eol_result' ->
+	    ( match best with
+		  EOL_found(pos',len') ->
+		    if pos < pos' || (pos = pos' && len > len') then
+		      find_best_eol r eol_result'
+		    else
+		      find_best_eol best eol_result'
+		| EOL_partially_found _ ->
+		    find_best_eol best eol_result'
+		| EOL_not_found ->
+		    find_best_eol r eol_result'
+	    )
+	| [] ->
+	    best
+    in
+    let eol_results = List.map find_this_eol eol in
+    find_best_eol EOL_not_found eol_results
+
+  method private enhanced_input s pos len : input_result =
+    if closed then raise Closed_channel;
+    if len > 0 then (
+      if bufpos = buflen then (
+	self # refill(); (* may raise End_of_file *)
+      );
+      let result = ref None in
+      while !result = None do
+	let best = self # find_eol() in
+	match best with
+	    EOL_not_found -> 
+	      let n = min len (buflen - bufpos) in
+	      String.blit buf bufpos s pos n;
+	      bufpos <- bufpos + n;
+	      result := Some(`Data n)
+	  | EOL_found(p,l) ->
+	      if p = bufpos then (
+		bufpos <- bufpos + l;
+		result := Some(`Separator(String.sub buf p l))
+	      )
+	      else (
+		let n = min len (p - bufpos) in
+		String.blit buf bufpos s pos n;
+		bufpos <- bufpos + n;
+		result := Some(`Data n)
+	      )
+	  | EOL_partially_found p ->
+	      if p = bufpos then (
+		try self # refill()
+		with End_of_file -> ()
+		(* ... and continue! *)
+	      )
+	      else (
+		let n = min len (p - bufpos) in
+		String.blit buf bufpos s pos n;
+		bufpos <- bufpos + n;
+		result := Some(`Data n)
+	      )
+      done;
+      match !result with
+	  None -> assert false
+	| Some r -> r
+    )
+    else `Data 0
+
+  method private enhanced_input_line() =
+    if closed then raise Closed_channel;
+    let b = Buffer.create 80 in
+    let eol_found = ref false in
+    if bufpos = buflen then (
+      self # refill();  (* may raise End_of_file *)
+    );
+    while not !eol_found do
+      let best = self # find_eol() in
+      try
+	match best with
+	    EOL_not_found ->
+	      Buffer.add_substring b buf bufpos (buflen-bufpos);
+	      bufpos <- buflen;
+	      self # refill();     (* may raise End_of_file *)
+	  | EOL_partially_found pos ->
+	      Buffer.add_substring b buf bufpos (pos-bufpos);
+	      bufpos <- pos;
+	      self # refill();     (* may raise End_of_file *)
+	  | EOL_found(pos,len) ->
+	      Buffer.add_substring b buf bufpos (pos-bufpos);
+	      bufpos <- pos+len;
+	      eol_found := true
+	with
+	    End_of_file -> 
+	      bufpos <- 0;
+	      buflen <- 0;
+	      eof <- true;
+	      eol_found := true
+    done;
+    Buffer.contents b
+end
+;;
+
+
+class lift_raw_in_channel_buf ?eol ?buffer_size r =
+object(self)
+  inherit buffered_raw_in_channel ?eol ?buffer_size r
+  inherit augment_raw_in_channel
+
+  method input_line () =
+    self # enhanced_input_line()
+end;;
+
+
+type lift_in_arg = [ `Rec of rec_in_channel | `Raw of raw_in_channel ]
+
+let lift_in ?(eol = ["\n"]) ?(buffered=true) ?buffer_size (x : lift_in_arg) =
+  match x with
+      `Rec r when not buffered ->
+	if eol <> ["\n"] then invalid_arg "Netchannels.lift_in";
+	new lift_rec_in_channel r
+    | `Rec r when buffered ->
+	let r' = new lift_rec_in_channel r in
+	new lift_raw_in_channel_buf ~eol ?buffer_size (r' :> raw_in_channel)
+    | `Raw r when not buffered ->
+	if eol <> ["\n"] then invalid_arg "Netchannels.lift_in";
+	new lift_raw_in_channel r
+    | `Raw r when buffered ->
+	new lift_raw_in_channel_buf ~eol ?buffer_size r
+;;
+
+
 (****************************** output ******************************)
 
 exception No_end_of_file
@@ -469,7 +819,7 @@ let copy_channel
       let m = min slen (match len with Some x -> x - !k | None -> max_int) in
       if m <= 0 then raise No_end_of_file;
       let n = src_ch # input buf 0 m in
-      if n = 0 then raise End_of_file;
+      if n = 0 then raise Sys_blocked_io;
       dest_ch # really_output buf 0 n;
       k := !k + n
     done;
@@ -494,8 +844,17 @@ object (self)
 
   method output buf pos len =
     if closed then self # complain_closed();
-    Pervasives.output ch buf pos len;
-    len
+    (* Pervasives.output does not support non-blocking I/O directly.
+     * Work around it:
+     *)
+    let p0 = Pervasives.pos_out ch in
+    try 
+      Pervasives.output ch buf pos len;
+      len
+    with
+	Sys_blocked_io ->
+	  let p1 = Pervasives.pos_out ch in
+	  p1 - p0
 
   method really_output buf pos len =
     if closed then self # complain_closed();
@@ -533,8 +892,6 @@ object (self)
     if closed then self # complain_closed();
     Pervasives.pos_out ch
 
-  method _rep_out = `Chan ch
-
 end
 ;;
 
@@ -607,7 +964,6 @@ object(self)
     if closed then self # complain_closed();
     Buffer.length buffer
 
-  method _rep_out = `Other
 end
 ;;
 
@@ -675,8 +1031,6 @@ object(self)
      * ever appended to the netbuffer.
      * This behavior is especially needed by [pipe_channel] below.
      *)
-
-  method _rep_out = `Other
 end
 ;;
 
@@ -716,7 +1070,6 @@ object(self)
   method pos_out =
     if closed then self # complain_closed();
     pos
-  method _rep_out = `Other
 end ;;
 
 
@@ -743,6 +1096,7 @@ object (self)
     let rec print_rest n =
       if n < len then
 	let m = self # output s (pos+n) (len-n) in
+	if m=0 then raise Sys_blocked_io;
 	print_rest (n+m)
       else
 	()
@@ -764,203 +1118,56 @@ object (self)
   method output_channel ?len ch =
     ignore(copy_channel ?len ch (self : #out_obj_channel :> out_obj_channel))
 
-  method _rep_out = `Other
 end
 ;;
 
 
-(************************* raw channels *******************************)
+class lift_raw_out_channel (r : raw_out_channel) =
+object(self)
+  inherit augment_raw_out_channel
 
-class input_descr ?(start_pos_in = 0) fd : raw_in_channel =
-object (self)
-  val fd = fd
-  val mutable pos_in = start_pos_in
-  val mutable closed = false
-
-  method private complain_closed() =
-    raise Closed_channel
-
-  method input buf pos len =
-    if closed then self # complain_closed();
-    try
-      let n = Unix.read fd buf pos len in
-      pos_in <- pos_in + n;
-      n
-    with
-	Unix.Unix_error(Unix.EINTR,_,_) ->
-	  self # input buf pos len
-  
-  method close_in () =
-    if closed then self # complain_closed();
-    Unix.close fd; closed <- true
-
-  method pos_in =
-    if closed then self # complain_closed();
-    pos_in
-end
-;;
-
-class output_descr ?(start_pos_out = 0) fd : raw_out_channel =
-object (self)
-  val fd = fd
-  val mutable pos_out = start_pos_out
-  val mutable closed = false
-
-  method private complain_closed() =
-    raise Closed_channel
-
-  method output buf pos len =
-    if closed then self # complain_closed();
-    try
-      let n = Unix.write fd buf pos len in
-      pos_out <- pos_out + n;
-      n
-    with
-	Unix.Unix_error(Unix.EINTR,_,_) ->
-	  self # output buf pos len
-  
-  method close_out () =
-    if closed then self # complain_closed();
-    Unix.close fd; closed <- true
-
-  method pos_out =
-    if closed then self # complain_closed();
-    pos_out
+  method output s p l =
+    r # output s p l
 
   method flush () =
-    if closed then self # complain_closed()
-end
-;;
+    r # flush()
 
-class socket_descr ?(start_pos_in = 0) ?(start_pos_out = 0) fd 
-      : raw_io_channel =
-object (self)
-  val fd = fd
-  val mutable pos_in = start_pos_in
-  val mutable pos_out = start_pos_out
-  val mutable in_closed = false
-  val mutable out_closed = false
-
-  method private complain_closed() =
-    raise Closed_channel
-
-  method input buf pos len =
-    if in_closed then self # complain_closed();
-    try
-      let n = Unix.read fd buf pos len in
-      pos_in <- pos_in + n;
-      n
-    with
-	Unix.Unix_error(Unix.EINTR,_,_) ->
-	  self # input buf pos len
-  
-  method close_in () =
-    if in_closed then self # complain_closed();
-    in_closed <- true;
-    Unix.shutdown fd Unix.SHUTDOWN_RECEIVE;
-    if out_closed then
-      Unix.close fd
-
-  method pos_in =
-    if in_closed then self # complain_closed();
-    pos_in
-
-  method output buf pos len =
-    if out_closed then self # complain_closed();
-    try
-      let n = Unix.write fd buf pos len in
-      pos_out <- pos_out + n;
-      n
-    with
-	Unix.Unix_error(Unix.EINTR,_,_) ->
-	  self # output buf pos len
-  
   method close_out () =
-    if out_closed then self # complain_closed();
-    out_closed <- true;
-    Unix.shutdown fd Unix.SHUTDOWN_SEND;
-    if in_closed then
-      Unix.close fd
+    r # close_out()
 
   method pos_out =
-    if out_closed then self # complain_closed();
-    pos_out
+    r # pos_out
+
+end;;
+
+
+class lift_rec_out_channel ?(start_pos_out = 0) (r : rec_out_channel) =
+object(self)
+  inherit augment_raw_out_channel
+
+  val mutable closed = false
+  val mutable pos_out = start_pos_out
+
+  method output s p l =
+    if closed then raise Closed_channel;
+    let n = r # output s p l in
+    pos_out <- pos_out + n;
+    n
 
   method flush() =
-    if out_closed then self # complain_closed()
-end
-;;
-
-
-class buffered_raw_in_channel
-      ?(buffer_size = 4096)
-      (ch : raw_in_channel) : enhanced_raw_in_channel =
-object (self)
-  val out = ch
-  val bufsize = buffer_size
-  val buf = String.create buffer_size
-  val mutable bufpos = 0
-  val mutable buflen = 0
-  val mutable eof = false
-  val mutable closed = false
-
-  method input s pos len =
     if closed then raise Closed_channel;
-    if bufpos = buflen then (
-      let n = ch # input buf 0 bufsize in
-      eof <- n = 0;
-      bufpos <- 0;
-      buflen <- n
-    );
-    if eof then 0 else (
-      let n = min len (buflen - bufpos) in
-      String.blit buf bufpos s pos n;
-      bufpos <- bufpos + n;
-      n
-    )
+    r # flush();
 
-  method close_in () =
-    ch # close_in();
-    closed <- true
+  method close_out () =
+    if closed then raise Closed_channel;
+    closed <- true;
+    r # close_out()
 
-  method pos_in =
-    (ch # pos_in) - (buflen - bufpos)
+  method pos_out =
+    if closed then raise Closed_channel;
+    pos_out
 
-  method private enhanced_input_line() =
-    let find_eol() =
-      let k =
-	try
-	  String.index_from buf bufpos '\n'
-	with
-	    Not_found -> buflen
-      in
-      min k buflen
-    in
-    let refill () =
-      let n = ch # input buf 0 bufsize in
-      eof <- n = 0;
-      bufpos <- 0;
-      buflen <- n
-    in
-    let b = Buffer.create 80 in
-    let eol_found = ref false in
-    if bufpos = buflen then (
-      refill();
-      if eof then raise End_of_file;
-    );
-    while not !eol_found do
-      let k = find_eol() in
-      Buffer.add_substring b buf bufpos (k-bufpos);
-      eol_found := (k < buflen);
-      bufpos <- k+1;
-      if bufpos > buflen then (
-	refill();
-	eol_found := eof;
-      )
-    done;
-    Buffer.contents b
-end
-;;
+end;;
 
 
 class buffered_raw_out_channel 
@@ -1001,6 +1208,129 @@ end
 ;;
 
 
+type lift_out_arg = [ `Rec of rec_out_channel | `Raw of raw_out_channel ]
+
+let lift_out ?(buffered=true) ?buffer_size (x : lift_out_arg) =
+  match x with
+      `Rec r when not buffered ->
+	new lift_rec_out_channel r
+    | `Rec r when buffered ->
+	let r' = new lift_rec_out_channel r in
+	let r'' = new buffered_raw_out_channel ?buffer_size (r' :> raw_out_channel) in
+	new lift_raw_out_channel r''
+    | `Raw r when not buffered ->
+	new lift_raw_out_channel r
+    | `Raw r when buffered ->
+	let r' = new buffered_raw_out_channel ?buffer_size r in
+	new lift_raw_out_channel r'
+;;
+
+
+(************************* raw channels *******************************)
+
+class input_descr_prelim ?(start_pos_in = 0) fd =
+object (self)
+  val fd_in = fd
+  val mutable pos_in = start_pos_in
+  val mutable closed_in = false
+
+  method private complain_closed() =
+    raise Closed_channel
+
+  method input buf pos len =
+    if closed_in then self # complain_closed();
+    try
+      let n = Unix.read fd_in buf pos len in
+      pos_in <- pos_in + n;
+      if n=0 && len>0 then raise End_of_file;
+      n
+    with
+	Unix.Unix_error(Unix.EINTR,_,_) ->
+	  self # input buf pos len
+      | Unix.Unix_error(Unix.EAGAIN,_,_)
+      | Unix.Unix_error(Unix.EWOULDBLOCK,_,_) ->
+	  0
+  
+  method close_in () =
+    if closed_in then self # complain_closed();
+    Unix.close fd_in; closed_in <- true
+
+  method pos_in =
+    if closed_in then self # complain_closed();
+    pos_in
+end
+;;
+
+
+class input_descr ?start_pos_in fd : raw_in_channel = 
+  input_descr_prelim ?start_pos_in fd
+;;
+
+
+class output_descr_prelim ?(start_pos_out = 0) fd =
+object (self)
+  val fd_out = fd
+  val mutable pos_out = start_pos_out
+  val mutable closed_out = false
+
+  method private complain_closed() =
+    raise Closed_channel
+
+  method output buf pos len =
+    if closed_out then self # complain_closed();
+    try
+      let n = Unix.write fd_out buf pos len in
+      pos_out <- pos_out + n;
+      n
+    with
+	Unix.Unix_error(Unix.EINTR,_,_) ->
+	  self # output buf pos len
+      | Unix.Unix_error(Unix.EAGAIN,_,_)
+      | Unix.Unix_error(Unix.EWOULDBLOCK,_,_) ->
+	  0
+  
+  method close_out () =
+    if closed_out then self # complain_closed();
+    Unix.close fd_out; closed_out <- true
+
+  method pos_out =
+    if closed_out then self # complain_closed();
+    pos_out
+
+  method flush () =
+    if closed_out then self # complain_closed()
+end
+;;
+
+
+class output_descr ?start_pos_out fd : raw_out_channel =
+  output_descr_prelim ?start_pos_out fd 
+;;
+
+
+class socket_descr ?(start_pos_in = 0) ?(start_pos_out = 0) fd 
+      : raw_io_channel =
+object (self)
+  inherit input_descr_prelim ~start_pos_in fd
+  inherit output_descr_prelim ~start_pos_out fd 
+  
+  method close_in () =
+    if closed_in then self # complain_closed();
+    closed_in <- true;
+    Unix.shutdown fd Unix.SHUTDOWN_RECEIVE;
+    if closed_out then
+      Unix.close fd
+  
+  method close_out () =
+    if closed_out then self # complain_closed();
+    closed_out <- true;
+    Unix.shutdown fd Unix.SHUTDOWN_SEND;
+    if closed_in then
+      Unix.close fd
+end
+;;
+
+
 (************************** transactional *****************************)
 
 class buffered_trans_channel ?(close_mode = (`Commit : close_mode)) 
@@ -1047,8 +1377,6 @@ object (self)
   method rollback_work() =
     Buffer.reset transbuf
 
-  method _rep_out = `Other
-
 end
 ;;
 
@@ -1194,9 +1522,6 @@ object (self)
      * underlying file descriptor will not be closed in this case
      *)
 
-
-  method _rep_out = `Other
-
 end
 ;;
 
@@ -1274,8 +1599,6 @@ object(self)
 
   method pos_in = input_super#pos_in
 
-  method _rep_in = input_super#_rep_in
-
   (* [close_out] also shuts down the input side of the pipe. *)
 
   method close_out () =
@@ -1345,8 +1668,6 @@ object(self)
 
   method pos_out = p # pos_out
 
-  method _rep_out = `Other
-
   method private transfer() =
     (* Copy as much as possible from [p] to [out] *)
     try
@@ -1408,14 +1729,27 @@ object(self)
 
   method pos_in = p#pos_in
 
-  method _rep_in = `Other
-
 end
 
 (* ======================================================================
  * History:
  * 
  * $Log: netchannels.ml,v $
+ * Revision 1.18  2004/07/11 15:40:29  stolpmann
+ * 	Added: enhanced_input
+ *
+ * Revision 1.17  2004/07/10 23:42:36  stolpmann
+ * 	New: delegation classes
+ * 	New: Arg eol of lift_in
+ * 	ocamldoc
+ *
+ * Revision 1.16  2004/05/30 21:27:14  stolpmann
+ * 	Supporting the standard for class-based I/O. This also means
+ * the semantics of [input] and [output] methods have changed (return
+ * value 0, End_of_file).
+ * 	Removed _rep_in, _rep_out.
+ * 	Removed Netstream.Deprecated.
+ *
  * Revision 1.15  2002/11/01 17:11:07  stolpmann
  * 	Changed EAGAIN -> EINTR
  *
