@@ -1,4 +1,4 @@
-(* $Id: netencoding.ml,v 2.10 2003/06/04 19:21:52 stolpmann Exp $
+(* $Id: netencoding.ml,v 2.15 2004/07/31 16:00:54 stolpmann Exp $
  * ----------------------------------------------------------------------
  *
  *)
@@ -451,45 +451,71 @@ end
 
 module QuotedPrintable = struct
 
-  let encode_substring s ~pos ~len =
+  let encode_substring ?(eot = false) ?(line_length = ref 0) s ~pos ~len =
+    (* line_length:
+     * - on input, the length of the line where the encoding starts
+     * - on output, the length of the last written line
+     * eot:
+     * - false: it is known that the chunk is not at the end of text
+     * - true: the chunk may be at the end of the text
+     * eot has only an effect on trailing spaces
+     *)
     
     if len < 0 or pos < 0 or pos > String.length s then
       invalid_arg "Netencoding.QuotedPrintable.encode";
     if pos + len > String.length s then
       invalid_arg "Netencoding.QuotedPrintable.encode";
 
-    let rec count n i =
+    (* Note: The [count] algorithm must strictly correspond to the
+     * "for" loop below.
+     *)
+
+    let rec count l n i =
+      (* l: output line length
+       * n: output byte count
+       * i: input byte count
+       *)
       if i < len then
 	match String.unsafe_get s (pos+i) with
-	    ('\r'|'\n') -> 
-	      count (n+1) (i+1)
+	    '\r' ->              (* CR is deleted *)
+	      count l n (i+1)
+	  | '\n' ->              (* LF is expanded to CR/LF *)
+	      count 0 (n+2) (i+1)
 	  | ('\000'..'\031'|'\127'..'\255'|
 	     '!'|'"'|'#'|'$'|'@'|'['|']'|'^'|'\''|'{'|'|'|'}'|'~'|'=') ->
-	      count (n+3) (i+1)
-	  | 'F' ->
+	      if l <= 69 then
+		count (l+3) (n+3) (i+1)
+	      else
+		(* Add soft line break after the encoded char: *)
+		count 0 (n+6) (i+1)
+	  | 'F' when l=0 ->
 	      (* Protect 'F' at the beginning of lines *)
-	      if i=0 || s.[pos+i-1]='\n' then
-		count (n+3) (i+1)
+	      count (l+3) (n+3) (i+1)
+	  | ' ' when (i=len-1 && eot) ||   (* at end of text *)
+	             l>69 ||               (* line too long *)
+                     (i<len-1 && (s.[pos+i+1]='\r' || s.[pos+i+1]='\n')) 
+		        (* end of line *)
+		     ->
+	      (* Protect spaces only if they occur at the end of a line,
+	       * or just before soft line breaks
+	       *)
+	      if l <= 69 then
+		count (l+3) (n+3) (i+1)
 	      else
-		count (n+1) (i+1)
-	  | ' ' ->
-	      (* Protect spaces only if they occur at the end of a line *)
-	      if i+1 < len then
-		match s.[pos+i+1] with
-		    ('\r'|'\n') -> 
-		      count (n+3) (i+1)
-		  | _ ->
-		      count (n+1) (i+1)
-	      else
-		count (n+3) (i+1)
+		(* Add soft line after the encoded space: *)
+		count 0 (n+6) (i+1)
 	  | _ ->
-	      count (n+1) (i+1)
+	      if l>71 then
+		(* Add soft line break after the char: *)
+		count 0 (n+4) (i+1)
+	      else
+		count (l+1) (n+1) (i+1)
       else
 	n
     in
 
-    let l = count 0 0 in
-    let t = String.create l in
+    let t_len = count !line_length 0 0 in
+    let t = String.create t_len in
     
     let hexdigit =
       [| '0'; '1'; '2'; '3'; '4'; '5'; '6'; '7';
@@ -503,53 +529,80 @@ module QuotedPrintable = struct
       t.[ !k+2 ] <- hexdigit.( Char.code c land 15 )
     in
 
+    let add_soft_break() =
+      t.[ !k ]   <- '=';
+      t.[ !k+1 ] <- '\r';
+      t.[ !k+2 ] <- '\n';
+    in
+
+    (* In the following, the soft break criterion is [!l > 72]. Why?
+     * We need to be able to add at least an encoded char (3 bytes)
+     * plus the "=" sign for the soft break. So we are on the safe side
+     * when there are four bytes space on the line. Lines must not be
+     * longer than 76 chars (w/o CRLF), so 76-4=72.
+     *)
+
+    let l = ref !line_length in
     for i = 0 to len - 1 do
       match String.unsafe_get s i with
-	  ('\r'|'\n') as c -> 
-	    String.unsafe_set t !k c;
-	    incr k
+	  '\r' ->   (* CR is deleted *)
+	    ()
+	| '\n' ->   (* LF is expanded to CR/LF *)
+	    t.[ !k ] <- '\r';
+	    t.[ !k+1 ] <- '\n';
+	    k := !k + 2;
+	    l := 0
 	| ('\000'..'\031'|'\127'..'\255'|
 	   '!'|'"'|'#'|'$'|'@'|'['|']'|'^'|'\''|'{'|'|'|'}'|'~'|'=') as c ->
 	    add_quoted c;
-	    k := !k + 3
-	| 'F' ->
-	    (* Protect 'F' at the beginning of lines *)
-	    if i=0 || s.[pos+i-1] = '\n' then begin
-	      add_quoted 'F';
-	      k := !k + 3
-	    end
-	    else begin
-	      String.unsafe_set t !k 'F';
-	      incr k
-	    end
-	| ' ' ->
-	    (* Protect spaces only if they occur at the end of a line *)
-	    if i+1 < len then
-	      match s.[pos+i+1] with
-		  ('\r'|'\n') -> 
-		    add_quoted ' ';
-		    k := !k + 3;
-		| _ ->
-		    String.unsafe_set t !k ' ';
-		    incr k
-	    else begin
-	      add_quoted ' ';
+	    k := !k + 3;
+	    l := !l + 3;
+	    if !l > 72 then (
+	      (* Add soft line break: *)
+	      add_soft_break();
 	      k := !k + 3;
-	    end
+	      l := 0
+	    )
+	| 'F' when !l = 0 ->
+	    (* Protect 'F' at the beginning of lines *)
+	    add_quoted 'F';
+	    k := !k + 3;
+	    l := !l + 3;
+	| ' ' when ((i=len-1 && eot) ||
+	              !l > 69 ||
+                     (i<len-1 && (s.[pos+i+1]='\r' || s.[pos+i+1]='\n'))) ->
+	    add_quoted ' ';
+	    k := !k + 3;
+	    l := !l + 3;
+	    if !l > 72 then (
+	      add_soft_break();
+	      k := !k + 3;
+	      l := 0;
+	    )
 	| c ->
 	    String.unsafe_set t !k c;
-	    incr k
+	    incr k;
+	    incr l;
+	    if !l > 72 then (
+	      add_soft_break();
+	      k := !k + 3;
+	      l := 0;
+	    )
     done;
 
+    assert(!k == t_len);
+
+    line_length := !l;
+
     t ;;
 
 
   let encode ?(pos=0) ?len s =
     let l = match len with None -> String.length s - pos | Some x -> x in 
-    encode_substring s pos l;;
+    encode_substring ~eot:true s ~pos ~len:l;;
 
 
-  let encoding_pipe_conv incoming incoming_eof outgoing =
+  let encoding_pipe_conv line_length incoming incoming_eof outgoing =
     (* Problematic case: the incoming buffer ends with a space, but we are
      * not at EOF. It is possible that a LF immediately follows, and that
      * the space needs to be quoted.
@@ -557,20 +610,21 @@ module QuotedPrintable = struct
      *)
     let s = Netbuffer.unsafe_buffer incoming in
     let len = Netbuffer.length incoming in
-    let len' =
+    let (len',eot) =
       if not incoming_eof && len > 0 && s.[len-1] = ' ' then
-	len-1
+	(len-1, false)
       else
-	len
+	(len, true)
     in
-    let s' = encode ~len:len' s in
+    let s' = encode_substring ~eot ~line_length s ~pos:0 ~len:len' in
     Netbuffer.add_string outgoing s';
     Netbuffer.delete incoming 0 len'
   ;;
     
 
   class encoding_pipe () =
-    Netchannels.pipe ~conv:encoding_pipe_conv ()
+    let line_length = ref 0 in
+    Netchannels.pipe ~conv:(encoding_pipe_conv line_length) ()
 
 
   let decode_substring s ~pos ~len =
@@ -703,7 +757,7 @@ module QuotedPrintable = struct
 
 
   module Deprecated = struct
-    let encode_substring = encode_substring
+    let encode_substring = encode_substring ?line_length:None
     let decode_substring = decode_substring
   end
 end
@@ -860,7 +914,7 @@ module Url = struct
 
 
   let url_encoding_re =
-    Str.regexp "[^A-Za-z0-9$_.!*'(),-]";;
+    Str.regexp "[^A-Za-z0-9_.!*-]";;
 
   let url_decoding_re =
     Str.regexp "\\+|%..|%.|%";;
@@ -1256,7 +1310,8 @@ module Html = struct
    *)
 
   let no_alphanum_re = Netstring_pcre.regexp "[^a-zA-Z0-9]";;
-  let null_re = Netstring_pcre.regexp "\\000";;
+  let null_re = Netstring_pcre.regexp "\\\\\\000";;
+      (* matches: backslash NUL *)
 
   let encode_quickly ~prefer_name ~unsafe_chars () =
     (* Preconditions: in_enc = out_enc, and the encoding must be a single-byte,
@@ -1267,6 +1322,7 @@ module Html = struct
       "]" in
     let unsafe_re_str' =
       Netstring_pcre.global_replace null_re "\\000" unsafe_re_str in
+      (* Replace backslash-NUL by PCRE null *)
     let unsafe_re = 
       Netstring_pcre.regexp unsafe_re_str' in
     Netstring_pcre.global_substitute
@@ -1292,6 +1348,7 @@ module Html = struct
       "\128-\255]" in
     let unsafe_re_str' =
       Netstring_pcre.global_replace null_re "\\000" unsafe_re_str in
+      (* Replace backslash-NUL by PCRE null *)
     let unsafe_re = 
       Netstring_pcre.regexp unsafe_re_str' in
     (* unicode_of.[q] = p: the code point q+128 of in_enc is the same as the
@@ -1516,6 +1573,25 @@ end
  * History:
  * 
  * $Log: netencoding.ml,v $
+ * Revision 2.15  2004/07/31 16:00:54  stolpmann
+ * 	Bugfix QuotedPrintable: Space at end of string no longer causes
+ * exception
+ *
+ * Revision 2.14  2004/07/18 16:36:24  stolpmann
+ * 	New: Neturl.file_url_of_local_path, local_path_of_file_url.
+ * 	URL updates according to RFC 2396.
+ *
+ * Revision 2.13  2004/07/11 15:51:18  stolpmann
+ * 	Minor bugfix in QuotedPrintable.encoding_pipe: Spaces at the
+ * end of the chunk but not at the end of the next are no longer
+ * quoted.
+ *
+ * Revision 2.12  2004/07/04 21:02:37  stolpmann
+ * 	Bug fix: QuotedPrintable takes care of adding soft line breaks.
+ *
+ * Revision 2.11  2004/05/30 21:53:08  stolpmann
+ * 	Fix: Html.encode no longer encodes "0"
+ *
  * Revision 2.10  2003/06/04 19:21:52  stolpmann
  * 	compile fixes
  *
