@@ -1,4 +1,4 @@
-(* $Id: nethtml.ml,v 2.6 2003/03/02 13:09:28 stolpmann Exp $
+(* $Id: nethtml.ml,v 2.8 2004/07/08 23:04:25 stolpmann Exp $
  * ----------------------------------------------------------------------
  *
  *)
@@ -364,6 +364,9 @@ let parse_document ?(dtd = html40_dtd)
     let backup_excl = !current_excl in
     try
       while not (is_possible_subelement !current_name !current_excl sub_name) do
+	(* Maybe we are not allowed to end the current element: *)
+	let (current_class, _) = model_of !current_name in
+	if current_class = `Essential_block then raise Stack.Empty;
 	(* End the current element and remove it from the stack: *)
 	let grant_parent = Stack.pop stack in
 	Stack.push grant_parent backup;        (* Save it; may we need it *)
@@ -371,8 +374,6 @@ let parse_document ?(dtd = html40_dtd)
 	(* If gp_name is an essential element, we are not allowed to close
 	 * it implicitly, even if that violates the DTD.
 	 *)
-	let (gp_class, _) = model_of gp_name in
-	if gp_class = `Essential_block then raise Stack.Empty;
 	let current = Element (!current_name, !current_atts, 
 			       List.rev !current_subs) in
 	current_name := gp_name;
@@ -706,36 +707,21 @@ let write ?(dtd = html40_dtd) ch doc =
   write_ ~dtd (ch # output_string) doc
 
 
-module Deprecated = struct
-let parse_string ?dtd ?return_declarations ?return_pis ?return_comments s =
-  let buf = Lexing.from_string s in
-  parse_document ?dtd ?return_declarations ?return_comments ?return_pis buf
-;;
-
-
-let parse_file ?dtd ?return_declarations ?return_pis ?return_comments fd =
-  let buf = Lexing.from_channel fd in
-  parse_document ?dtd ?return_declarations ?return_comments ?return_pis buf
-;;
-
-let write ?(dtd = html40_dtd) os doc =
-  let write_os =
-    match os with
-	`Out_buffer b ->
-	  (fun s -> Buffer.add_string b s)
-      | `Out_channel ch ->
-	  (fun s -> output_string ch s)
-      | `Out_function f ->
-	  (fun s -> f s 0 (String.length s))
-  in
-  write_ ~dtd write_os doc
-
-end
-
 (* ======================================================================
  * History:
  * 
  * $Log: nethtml.ml,v $
+ * Revision 2.8  2004/07/08 23:04:25  stolpmann
+ * 	ocamldoc
+ *
+ * Revision 2.7  2004/06/23 21:47:52  stolpmann
+ * 	Handling of essential blocks was wrong when the block was
+ * the direct parent of the element to add. The block was implicitly
+ * closed in this case although this is not allowed (e.g.
+ * <ul><li>x</li><ol><li>y</li></ol></ul>: When <ol> is found, the
+ * ul element must not be closed).
+ * 	This is now fixed.
+ *
  * Revision 2.6  2003/03/02 13:09:28  stolpmann
  * 	Fixed dl model
  *
