@@ -1,4 +1,4 @@
-(* $Id: netmime.ml,v 1.7 2002/06/09 11:35:16 stolpmann Exp $
+(* $Id: netmime.ml,v 1.9 2004/07/06 17:24:22 stolpmann Exp $
  * ----------------------------------------------------------------------
  *
  *)
@@ -84,6 +84,159 @@ end
 module CIMap = Map.Make(CI)
   (* Maps from case-insensitive strings to any type *)
 
+module DL : sig  (* doubly-linked lists *)
+  type 'a t
+  type 'a cell
+  val create : unit -> 'a t
+  val is_empty : 'a t -> bool
+  val cell : 'a -> 'a cell
+  val contents : 'a cell -> 'a
+  val first : 'a t -> 'a cell (* or Not_found *)
+  val last : 'a t -> 'a cell (* or Not_found *)
+  val prev : 'a cell -> 'a cell (* or Not_found *)
+  val next : 'a cell -> 'a cell (* or Not_found *)
+  val iter : ('a cell -> unit) -> 'a t -> unit
+  val delete : 'a cell -> unit
+  val insert_after : neo:'a cell -> 'a cell -> unit
+  val add_at_end : neo:'a cell -> 'a t -> unit
+  val replace : neo:'a cell -> 'a cell -> unit
+  val of_list : 'a list -> 'a t
+  val to_list : 'a t -> 'a list
+end = struct
+  type 'a t =
+      { mutable first : 'a cell option;
+	mutable last : 'a cell option;
+      }
+  and 'a cell =
+      { mutable prev : 'a cell option;
+	mutable next : 'a cell option;
+	mutable list : 'a t option;
+	contents : 'a;
+      }
+  let create() = 
+    { first = None; last = None }
+  let is_empty l =
+    l.first = None
+  let cell x =
+    { prev = None; next = None; list = None; contents = x }
+  let contents c =
+    c.contents
+  let first l =
+    match l.first with Some c -> c | None -> raise Not_found
+  let last l =
+    match l.last with Some c -> c | None -> raise Not_found
+  let prev c =
+    match c.prev with Some c' -> c' | None -> raise Not_found
+  let next c =
+    match c.next with Some c' -> c' | None -> raise Not_found
+  let iter f l =
+    match l.first with
+	Some c ->
+	  f c;
+	  let cur = ref c in
+	  while !cur.next <> None do
+	    cur := next !cur;
+	    f !cur
+	  done; ()
+      | None ->
+	  ()
+  let delete c =
+    match c.list with
+	Some l ->
+	  ( match c.prev with
+		Some p ->
+		  p.next <- c.next
+	      | None ->
+		  l.first <- c.next
+	  );
+	  ( match c.next with
+		Some n ->
+		  n.prev <- c.prev
+	      | None ->
+		  l.last <- c.prev
+	  );
+	  c.prev <- None;
+	  c.next <- None;
+	  c.list <- None
+      | None -> 
+	  failwith "DL.delete: cannot delete free cell"
+  let insert_after ~neo c =
+    if neo.list <> None then
+      failwith "DL.insert_after: new cell must be free";
+    match c.list with
+	Some l ->
+	  let nx = c.next in
+	  c.next <- Some neo;
+	  neo.prev <- Some c;
+	  ( match nx with
+		Some n ->
+		  n.prev <- Some neo;
+		  neo.next <- Some n;
+	      | None ->
+		  l.last <- Some neo;
+		  neo.next <- None
+	  );
+	  neo.list <- Some l
+      | None -> 
+	  failwith "DL.insert_after: cannot insert after free cell"
+  let add_at_end ~neo l =
+    if neo.list <> None then
+      failwith "DL.insert_after: new cell must be free";
+    match l.last with
+	Some n ->
+	  n.next <- Some neo;
+	  neo.prev <- Some n;
+	  neo.next <- None;
+	  neo.list <- Some l;
+	  l.last <- Some neo
+      | None ->
+	  l.last <- Some neo;
+	  l.first <- Some neo;
+	  neo.prev <- None;
+	  neo.next <- None;
+	  neo.list <- Some l
+  let replace ~neo c =
+    if neo.list <> None then
+      failwith "DL.replace: new cell must be free";
+    match c.list with
+	Some l ->
+	  ( match c.prev with
+		Some p ->
+		  p.next <- Some neo
+	      | None ->
+		  l.first <- Some neo
+	  );
+	  neo.prev <- c.prev;
+	  ( match c.next with
+		Some n ->
+		  n.prev <- Some neo
+	      | None ->
+		  l.last <- Some neo
+	  );
+	  neo.next <- c.next;
+	  neo.list <- Some l;
+	  c.prev <- None;
+	  c.next <- None;
+	  c.list <- None
+      | None -> 
+	  failwith "DL.replace: cannot replace free cell"
+  let of_list l =
+    let dl = create() in
+    List.iter
+      (fun x -> 
+	 add_at_end ~neo:(cell x) dl
+      )
+      l;
+    dl
+  let rec to_list dl =
+    chain_to_list dl.first
+  and chain_to_list chain =
+    match chain with
+	None -> []
+      | Some c -> c.contents :: chain_to_list c.next
+
+end
+
 
 let drop_ws_re = Netstring_pcre.regexp "^[ \t\r\n]*(.*[^ \t\r\n])[ \t\r\n]*$";;
 
@@ -100,8 +253,8 @@ let drop_ws s =
 class basic_mime_header ?(ro=false) h : mime_header =
 object (self)
   val ro = ro
-  val mutable hdr = lazy (assert false)
-  val mutable hdr_fields = None
+  val mutable hdr_map = lazy (assert false)
+  val mutable hdr_dl  = lazy (assert false)
 
   initializer
     self # do_set_fields h
@@ -109,23 +262,20 @@ object (self)
   method ro = ro
 
   method fields = 
-    match hdr_fields with
-	Some f -> f
-      | None ->
-	  let m,order = Lazy.force hdr in
-	  List.flatten
-	    (List.rev_map
-	       (fun n_ci -> CIMap.find n_ci m)
-	       order
-	    )
+    DL.to_list (Lazy.force hdr_dl)
 
   method field n = 
-    let m,order = Lazy.force hdr in
-    snd (List.hd (CIMap.find (CI.make n) m))
+    let m = Lazy.force hdr_map in
+    match CIMap.find (CI.make n) m with
+	[] -> raise Not_found
+      | cell :: _ -> snd (DL.contents cell)
 
   method multiple_field n =
-    let m,order = Lazy.force hdr in
-    try  List.map snd (CIMap.find (CI.make n) m) 
+    let m = Lazy.force hdr_map in
+    try 
+      List.map 
+        (fun cell -> snd (DL.contents cell)) 
+	(CIMap.find (CI.make n) m) 
     with Not_found -> []
 
   method private immutable() =
@@ -136,52 +286,91 @@ object (self)
     self # do_set_fields h
 
   method private do_set_fields h =
-    hdr <- lazy begin
+    hdr_dl <- lazy (DL.of_list h);
+    hdr_map <- lazy begin
       (* This seems to be expensive (O(n log n)). Because of this we do it only
        * on demand; maybe nobody accesses the header at all
        *)
       let m = ref CIMap.empty in
-      let order = ref [] in
-      List.iter
-	(fun (n,v) -> 
+      DL.iter
+	(fun cell ->
+	   let (n,v) = DL.contents cell in
 	   let n_ci = CI.make n in
 	   let current = 
 	     try CIMap.find n_ci !m 
-	     with Not_found -> order := n_ci :: !order; [] 
+	     with Not_found -> [] 
 	   in
-	   m := CIMap.add n_ci ((n,v) :: current) !m;
+	   m := CIMap.add n_ci (cell :: current) !m;
 	)
-	h;
-      (CIMap.map List.rev !m, !order)
-    end;
-    hdr_fields <- Some h
+	(Lazy.force hdr_dl);
+      CIMap.map List.rev !m
+    end
 
   method update_field n v =
-    if ro then self#immutable();
-    let n_ci = CI.make n in
-    let m, order = Lazy.force hdr in
-    let m' = CIMap.add n_ci [n,v] m in
-    let order' = if CIMap.mem n_ci m then order else n_ci :: order in
-    hdr <- lazy(m',order');
-    hdr_fields <- None
+    self # update_multiple_field n [v]
 
   method update_multiple_field n vl =
     if ro then self#immutable();
     let n_ci = CI.make n in
-    let m, order = Lazy.force hdr in
-    let m' = CIMap.add n_ci (List.map (fun v -> (n,v)) vl) m in
-    let order' = if CIMap.mem n_ci m then order else n_ci :: order in
-    hdr <- lazy(m',order');
-    hdr_fields <- None
+    let m = Lazy.force hdr_map in
+    let dl = Lazy.force hdr_dl in
+    (* Algorithm: First try to replace existing values.
+     * If there are more new values than old values,
+     * at the excess values after the last old value,
+     * or if not possible, at the end.
+     *)
+    let insert_point =
+      ref None in
+    let old_cells =
+      ref(try CIMap.find n_ci m with Not_found -> []) in
+    let new_vals  = ref vl in
+    let new_cells = ref [] in
+    while !old_cells <> [] || !new_vals <> [] do
+      match !old_cells, !new_vals with
+	  (old_cell :: old_cells'), (new_val :: new_vals') ->
+	    (* Only update if the value has changed: *)
+	    let (old_n, old_val) = DL.contents old_cell in
+	    if old_val = new_val then (
+	      new_cells := old_cell :: !new_cells;
+	      insert_point := Some old_cell;
+	    )
+	    else (
+	      let new_cell = DL.cell (n, new_val) in
+	      DL.replace ~neo:new_cell old_cell;
+	      insert_point := Some new_cell;
+	      new_cells := new_cell :: !new_cells
+	    );
+	    old_cells := old_cells';
+	    new_vals  := new_vals';
+	| [], (new_val :: new_vals') ->
+	    let new_cell = DL.cell (n, new_val) in
+	    ( match !insert_point with
+		  Some p ->
+		    DL.insert_after ~neo:new_cell p;
+		| None ->
+		    DL.add_at_end ~neo:new_cell dl
+	    );
+	    new_vals := new_vals';
+	    insert_point := Some new_cell;
+	    new_cells := new_cell :: !new_cells
+	| (old_cell :: old_cells'), [] ->
+	    DL.delete old_cell;
+	    old_cells := old_cells'
+	| [], [] ->
+	    assert false
+    done;
+    let m' = CIMap.add n_ci (List.rev !new_cells) m in
+    hdr_map <- lazy m'
 
   method delete_field n =
     if ro then self#immutable();
     let n_ci = CI.make n in
-    let m, order = Lazy.force hdr in
+    let m = Lazy.force hdr_map in
+    let old_cells =
+      try CIMap.find n_ci m with Not_found -> [] in
+    List.iter DL.delete old_cells;
     let m' = CIMap.remove n_ci m in
-    let order' = List.filter (fun n_ci' -> n_ci' <> n_ci) order in
-    hdr <- lazy(m',order');
-    hdr_fields <- None
+    hdr_map <- lazy m';
 
   method content_length() = 
     int_of_string (drop_ws(self # field "content-length"))
@@ -512,7 +701,11 @@ let rec write_mime_message_int ?(wr_header = true) ?(wr_body = true) ?(nr = 0)
 	    with_in_obj_channel
 	      (body # open_value_rd())
 	      (fun ch ->
-		 ignore(ch # input s 0 240));
+		 try 
+		   ignore(ch # input s 0 240)
+		 with
+		     End_of_file -> ()  (* When body is empty *)
+	      );
 	    [s]
 	| (_,`Parts_ag(_, parts'')) :: parts' ->
 	    (try gather_data parts'' with Not_found -> gather_data parts')
@@ -614,6 +807,14 @@ let write_mime_message ?wr_header ?wr_body ?nr ?ret_boundary ch msg =
  * History:
  * 
  * $Log: netmime.ml,v $
+ * Revision 1.9  2004/07/06 17:24:22  stolpmann
+ * 	basic_mime_header respects the order of fields even when fields
+ * with a certain name are not adjacent. Changed the impl of basic_mime_header
+ * to a doubly linked list, with a Map as dictionary to the cells of the list.
+ *
+ * Revision 1.8  2004/05/30 21:24:04  stolpmann
+ * 	Fixed bug in method [field] of [basic_mime_header]
+ *
  * Revision 1.7  2002/06/09 11:35:16  stolpmann
  * 	Passing ~strip as suggested by Matt Armstrong.
  *
