@@ -1,4 +1,4 @@
-(* $Id: neturl.ml,v 2.1 2001/09/14 14:22:34 stolpmann Exp $
+(* $Id: neturl.ml,v 2.3 2004/07/18 16:36:24 stolpmann Exp $
  * ----------------------------------------------------------------------
  *
  *)
@@ -14,6 +14,7 @@ type url_syntax_option =
 type url_syntax =
     { url_enable_scheme    : url_syntax_option;
       url_enable_user      : url_syntax_option;
+      url_enable_user_param : url_syntax_option;
       url_enable_password  : url_syntax_option;
       url_enable_host      : url_syntax_option;
       url_enable_port      : url_syntax_option;
@@ -24,6 +25,7 @@ type url_syntax =
       url_enable_other     : url_syntax_option;
       url_accepts_8bits    : bool;
       url_is_valid         : url -> bool;
+      url_enable_relative  : bool;
     }
 
 and url =
@@ -32,6 +34,7 @@ and url =
       mutable url_validity : bool;
       url_scheme   : string option;
       url_user     : string option;
+      url_user_param : string list;
       url_password : string option;
       url_host     : string option;
       url_port     : int option;
@@ -153,12 +156,16 @@ let scheme_cats =
 
 let login_cats =
   make_cats 
-    (lalpha @ ualpha @ digit @ safe @ extra @ [';'; '?'; '&'; '='; '%'])  
-    [':'; '@'; '/'; '#' ]
+    (lalpha @ ualpha @ digit @ safe @ extra @ [';'; '&'; '='; '%'])  
+    [':'; '@'; '/'; '#'; '?' ]
 ;;
 
     (* login_cats: character categorization to _extract_ user name, password,
      * host name, and port.
+     *
+     * Note: user_params are extracted in a second step.
+     *
+     * Note: '?' is now a separator, as described in RFC 2396.
      *)
 
 let host_cats =
@@ -170,6 +177,8 @@ let host_cats =
     (* host_cats: character categorization to _check_ whether the host name
      * is formed only by legal characters.
      * Especially '%' is not allowed here!
+     *
+     * CHECK: IPv6
      *)
 
 let port_cats =
@@ -216,7 +225,8 @@ let path_cats_from_syntax syn extraseps =
  *
  * Notes:
  * - The '#' is rejected unless fragments are enabled. 
- * - The '~' is accepted although this violates RFC 1738.
+ * - The '~' is accepted although this violates RFC 1738 (but it is ok
+ *   according to RFC 2396)
  *)
 
 
@@ -262,9 +272,10 @@ let ( <=> ) (a:bool) b = ( a = b );;  (* equivalence *)
 let url_syntax_is_valid syn =
   let recognized x = x <> Url_part_not_recognized in
   let not_recognized x = x = Url_part_not_recognized in
-  (recognized syn.url_enable_password => recognized syn.url_enable_user) &
-  (recognized syn.url_enable_port     => recognized syn.url_enable_host) &
-  (recognized syn.url_enable_user     => recognized syn.url_enable_host) &
+  (recognized syn.url_enable_password   => recognized syn.url_enable_user) &&
+  (recognized syn.url_enable_user_param => recognized syn.url_enable_user) &&
+  (recognized syn.url_enable_port       => recognized syn.url_enable_host) &&
+  (recognized syn.url_enable_user       => recognized syn.url_enable_host) &&
   not ( (recognized syn.url_enable_user ||
 	 recognized syn.url_enable_password ||
 	 recognized syn.url_enable_host ||
@@ -281,8 +292,11 @@ let partial_url_syntax syn =
       | Url_part_allowed        -> Url_part_allowed
       | Url_part_required       -> Url_part_allowed
   in
+  if not syn.url_enable_relative then
+    failwith "Neturl.partial_url_syntax: This syntax does not support relative URLs";
   { url_enable_scheme    = weaken syn.url_enable_scheme;
     url_enable_user      = weaken syn.url_enable_user;
+    url_enable_user_param= weaken syn.url_enable_user_param;
     url_enable_password  = weaken syn.url_enable_password;
     url_enable_host      = weaken syn.url_enable_host;
     url_enable_port      = weaken syn.url_enable_port;
@@ -293,6 +307,7 @@ let partial_url_syntax syn =
     url_enable_other     = weaken syn.url_enable_other;
     url_accepts_8bits    = syn.url_accepts_8bits;
     url_is_valid         = syn.url_is_valid;
+    url_enable_relative  = true;
   }
 ;;
 
@@ -301,6 +316,7 @@ let partial_url_syntax syn =
 let file_url_syntax =
   { url_enable_scheme    = Url_part_required;
     url_enable_user      = Url_part_not_recognized;
+    url_enable_user_param= Url_part_not_recognized;
     url_enable_password  = Url_part_not_recognized;
     url_enable_host      = Url_part_allowed;
     url_enable_port      = Url_part_not_recognized;
@@ -311,6 +327,7 @@ let file_url_syntax =
     url_enable_other     = Url_part_not_recognized;
     url_accepts_8bits    = false;
     url_is_valid         = (fun _ -> true);
+    url_enable_relative  = true;
   }
 ;;
 
@@ -318,6 +335,7 @@ let file_url_syntax =
 let ftp_url_syntax =
   { url_enable_scheme    = Url_part_required;
     url_enable_user      = Url_part_allowed;
+    url_enable_user_param= Url_part_not_recognized;
     url_enable_password  = Url_part_allowed;
     url_enable_host      = Url_part_required;
     url_enable_port      = Url_part_allowed;
@@ -328,6 +346,7 @@ let ftp_url_syntax =
     url_enable_other     = Url_part_not_recognized;
     url_accepts_8bits    = false;
     url_is_valid         = (fun _ -> true);
+    url_enable_relative  = true;
   }
 ;;
 
@@ -335,6 +354,7 @@ let ftp_url_syntax =
 let http_url_syntax =
   { url_enable_scheme    = Url_part_required;
     url_enable_user      = Url_part_allowed;
+    url_enable_user_param= Url_part_not_recognized;
     url_enable_password  = Url_part_allowed;
     url_enable_host      = Url_part_required;
     url_enable_port      = Url_part_allowed;
@@ -345,6 +365,45 @@ let http_url_syntax =
     url_enable_other     = Url_part_not_recognized;
     url_accepts_8bits    = false;
     url_is_valid         = (fun _ -> true);
+    url_enable_relative  = true;
+  }
+;;
+
+
+let pop_url_syntax =
+  { url_enable_scheme    = Url_part_required;
+    url_enable_user      = Url_part_allowed;
+    url_enable_user_param= Url_part_allowed;
+    url_enable_password  = Url_part_allowed;
+    url_enable_host      = Url_part_required;
+    url_enable_port      = Url_part_allowed;
+    url_enable_path      = Url_part_not_recognized;
+    url_enable_param     = Url_part_not_recognized;
+    url_enable_query     = Url_part_not_recognized;
+    url_enable_fragment  = Url_part_not_recognized;
+    url_enable_other     = Url_part_not_recognized;
+    url_accepts_8bits    = false;
+    url_is_valid         = (fun _ -> true);
+    url_enable_relative  = false;
+  }
+;;
+
+
+let imap_url_syntax =
+  { url_enable_scheme    = Url_part_required;
+    url_enable_user      = Url_part_allowed;
+    url_enable_user_param= Url_part_allowed;
+    url_enable_password  = Url_part_allowed;
+    url_enable_host      = Url_part_required;
+    url_enable_port      = Url_part_allowed;
+    url_enable_path      = Url_part_allowed;
+    url_enable_param     = Url_part_not_recognized;
+    url_enable_query     = Url_part_allowed;
+    url_enable_fragment  = Url_part_not_recognized;
+    url_enable_other     = Url_part_not_recognized;
+    url_accepts_8bits    = false;
+    url_is_valid         = (fun _ -> true);
+    url_enable_relative  = true;
   }
 ;;
 
@@ -352,6 +411,26 @@ let http_url_syntax =
 let mailto_url_syntax =
   { url_enable_scheme    = Url_part_required;
     url_enable_user      = Url_part_not_recognized;
+    url_enable_user_param= Url_part_not_recognized;
+    url_enable_password  = Url_part_not_recognized;
+    url_enable_host      = Url_part_not_recognized;
+    url_enable_port      = Url_part_not_recognized;
+    url_enable_path      = Url_part_not_recognized;
+    url_enable_param     = Url_part_not_recognized;
+    url_enable_query     = Url_part_allowed;
+    url_enable_fragment  = Url_part_not_recognized;
+    url_enable_other     = Url_part_required;
+    url_accepts_8bits    = false;
+    url_is_valid         = (fun _ -> true);
+    url_enable_relative  = false;
+  }
+;;
+
+
+let news_url_syntax =
+  { url_enable_scheme    = Url_part_required;
+    url_enable_user      = Url_part_not_recognized;
+    url_enable_user_param= Url_part_not_recognized;
     url_enable_password  = Url_part_not_recognized;
     url_enable_host      = Url_part_not_recognized;
     url_enable_port      = Url_part_not_recognized;
@@ -362,6 +441,48 @@ let mailto_url_syntax =
     url_enable_other     = Url_part_required;
     url_accepts_8bits    = false;
     url_is_valid         = (fun _ -> true);
+    url_enable_relative  = false;
+  }
+;;
+let cid_url_syntax = news_url_syntax;;
+let data_url_syntax = news_url_syntax;;
+
+let nntp_url_syntax =
+  { url_enable_scheme    = Url_part_required;
+    url_enable_user      = Url_part_not_recognized;
+    url_enable_user_param= Url_part_not_recognized;
+    url_enable_password  = Url_part_not_recognized;
+    url_enable_host      = Url_part_required;
+    url_enable_port      = Url_part_allowed;
+    url_enable_path      = Url_part_required;
+    url_enable_param     = Url_part_not_recognized;
+    url_enable_query     = Url_part_not_recognized;
+    url_enable_fragment  = Url_part_not_recognized;
+    url_enable_other     = Url_part_not_recognized;
+    url_accepts_8bits    = false;
+    url_is_valid         = (fun url -> 
+			      List.length url.url_path = 3
+			   );
+    url_enable_relative  = true;
+  }
+;;
+
+
+let ipp_url_syntax =
+  { url_enable_scheme    = Url_part_required;
+    url_enable_user      = Url_part_not_recognized;
+    url_enable_user_param= Url_part_not_recognized;
+    url_enable_password  = Url_part_not_recognized;
+    url_enable_host      = Url_part_required;
+    url_enable_port      = Url_part_allowed;
+    url_enable_path      = Url_part_allowed;
+    url_enable_param     = Url_part_not_recognized;
+    url_enable_query     = Url_part_allowed;
+    url_enable_fragment  = Url_part_not_recognized;
+    url_enable_other     = Url_part_not_recognized;
+    url_accepts_8bits    = false;
+    url_is_valid         = (fun _ -> true);
+    url_enable_relative  = true;
   }
 ;;
 
@@ -369,6 +490,7 @@ let mailto_url_syntax =
 let null_url_syntax =
   { url_enable_scheme    = Url_part_not_recognized;
     url_enable_user      = Url_part_not_recognized;
+    url_enable_user_param= Url_part_not_recognized;
     url_enable_password  = Url_part_not_recognized;
     url_enable_host      = Url_part_not_recognized;
     url_enable_port      = Url_part_not_recognized;
@@ -379,6 +501,7 @@ let null_url_syntax =
     url_enable_other     = Url_part_not_recognized;
     url_accepts_8bits    = false;
     url_is_valid         = (fun _ -> true);
+    url_enable_relative  = false;
   }
 ;;
 
@@ -386,6 +509,10 @@ let null_url_syntax =
 let ip_url_syntax =
   { url_enable_scheme    = Url_part_allowed;
     url_enable_user      = Url_part_allowed;
+    url_enable_user_param= Url_part_not_recognized;
+                           (* so user parameters are parsed as part
+			    * of the user string! (Most generic.)
+			    *)
     url_enable_password  = Url_part_allowed;
     url_enable_host      = Url_part_allowed;
     url_enable_port      = Url_part_allowed;
@@ -396,6 +523,7 @@ let ip_url_syntax =
     url_enable_other     = Url_part_not_recognized;
     url_accepts_8bits    = false;
     url_is_valid         = (fun _ -> true);
+    url_enable_relative  = true;
   }
 ;;
 
@@ -405,7 +533,20 @@ let common_url_syntax =
   Hashtbl.add h "file"   file_url_syntax;
   Hashtbl.add h "ftp"    ftp_url_syntax;
   Hashtbl.add h "http"   http_url_syntax;
+  Hashtbl.add h "https"  http_url_syntax;
+  Hashtbl.add h "pop"    pop_url_syntax;
+  Hashtbl.add h "pops"   pop_url_syntax;
+  Hashtbl.add h "imap"   imap_url_syntax;
+  Hashtbl.add h "imaps"  imap_url_syntax;
   Hashtbl.add h "mailto" mailto_url_syntax;
+  Hashtbl.add h "news"   news_url_syntax;
+  Hashtbl.add h "nntp"   nntp_url_syntax;
+  Hashtbl.add h "nntps"  nntp_url_syntax;
+  Hashtbl.add h "cid"    cid_url_syntax;
+  Hashtbl.add h "mid"    cid_url_syntax;
+  Hashtbl.add h "data"   data_url_syntax;
+  Hashtbl.add h "ipp"    ipp_url_syntax;
+  Hashtbl.add h "ipps"   ipp_url_syntax;
   h
 ;;
 
@@ -415,27 +556,29 @@ let url_conforms_to_syntax url =
   let required x = x = Url_part_required in
   let present x    = x <> None in
   let syn = url.url_syntax in
-  (present url.url_scheme   => recognized syn.url_enable_scheme)   &
-  (present url.url_user     => recognized syn.url_enable_user)     &
-  (present url.url_password => recognized syn.url_enable_password) &
-  (present url.url_host     => recognized syn.url_enable_host)     &
-  (present url.url_port     => recognized syn.url_enable_port)     &
-  ((url.url_path <> [])     => recognized syn.url_enable_path)     &
-  ((url.url_param <> [])    => recognized syn.url_enable_param)    &
-  (present url.url_query    => recognized syn.url_enable_query)    &
-  (present url.url_fragment => recognized syn.url_enable_fragment) &
-  (present url.url_other    => recognized syn.url_enable_other)    &
-  (required syn.url_enable_scheme   => present url.url_scheme)     &
-  (required syn.url_enable_user     => present url.url_user)       &
-  (required syn.url_enable_password => present url.url_password)   &
-  (required syn.url_enable_host     => present url.url_host)       &
-  (required syn.url_enable_port     => present url.url_port)       &
-  (required syn.url_enable_path     => (url.url_path <> []))       &
-  (required syn.url_enable_param    => (url.url_param <> []))      &
-  (required syn.url_enable_query    => present url.url_query)      &
-  (required syn.url_enable_fragment => present url.url_fragment)   &
-  (required syn.url_enable_other    => present url.url_other)      &
-  (url.url_validity or syn.url_is_valid url)
+  (present url.url_scheme   => recognized syn.url_enable_scheme)   &&
+  (present url.url_user     => recognized syn.url_enable_user)     &&
+  ((url.url_user_param <> []) => recognized syn.url_enable_user_param) &&
+  (present url.url_password => recognized syn.url_enable_password) &&
+  (present url.url_host     => recognized syn.url_enable_host)     &&
+  (present url.url_port     => recognized syn.url_enable_port)     &&
+  ((url.url_path <> [])     => recognized syn.url_enable_path)     &&
+  ((url.url_param <> [])    => recognized syn.url_enable_param)    &&
+  (present url.url_query    => recognized syn.url_enable_query)    &&
+  (present url.url_fragment => recognized syn.url_enable_fragment) &&
+  (present url.url_other    => recognized syn.url_enable_other)    &&
+  (required syn.url_enable_scheme   => present url.url_scheme)     &&
+  (required syn.url_enable_user     => present url.url_user)       &&
+  (required syn.url_enable_user_param => (url.url_user_param <> [])) &&
+  (required syn.url_enable_password => present url.url_password)   &&
+  (required syn.url_enable_host     => present url.url_host)       &&
+  (required syn.url_enable_port     => present url.url_port)       &&
+  (required syn.url_enable_path     => (url.url_path <> []))       &&
+  (required syn.url_enable_param    => (url.url_param <> []))      &&
+  (required syn.url_enable_query    => present url.url_query)      &&
+  (required syn.url_enable_fragment => present url.url_fragment)   &&
+  (required syn.url_enable_other    => present url.url_other)      &&
+  (url.url_validity || syn.url_is_valid url)
 ;;
 
 
@@ -448,6 +591,7 @@ let modify_url
       ?(encoded = false)
       ?scheme
       ?user
+      ?user_param
       ?password
       ?host
       ?port
@@ -511,6 +655,9 @@ let modify_url
       url_validity = false;
       url_scheme   = if scheme   = None then url.url_scheme   else scheme;
       url_user     = if user     = None then url.url_user     else enc user;
+      url_user_param = ( match user_param with 
+			     None -> url.url_user_param
+			   | Some p -> enc_list p);
       url_password = if password = None then url.url_password else enc password;
       url_host     = if host     = None then url.url_host     else host;
       url_port     = if port     = None then url.url_port     else port;
@@ -529,13 +676,15 @@ let modify_url
    *)
   if not (url_conforms_to_syntax url') then raise Malformed_URL;
   if url'.url_password <> None && url'.url_user = None then raise Malformed_URL;
+  if url'.url_user_param <> [] && url'.url_user = None then raise Malformed_URL;
   if url'.url_user <> None && url'.url_host = None then raise Malformed_URL;
   if url'.url_port <> None && url'.url_host = None then raise Malformed_URL;
   (* Check every part: *)
-  check_string url'.url_scheme   scheme_cats;
-  check_string url'.url_user     login_cats;
-  check_string url'.url_password login_cats;
-  check_string url'.url_host     host_cats;
+  check_string url'.url_scheme          scheme_cats;
+  check_string url'.url_user            login_cats;
+  check_string_list url'.url_user_param login_cats ';';
+  check_string url'.url_password        login_cats;
+  check_string url'.url_host            host_cats;
   (match url'.url_port with 
        None -> ()
      | Some p -> if p < 0 || p > 65535 then raise Malformed_URL
@@ -551,10 +700,12 @@ let modify_url
   (* Further path checks: *)
   begin match url'.url_path with
       [] ->
-	(* The path is empty: There must not be a 'param' or 'query' *)
+	(* The path is empty: There must not be 'param' *)
 	if url'.url_host <> None then begin
 	  if url'.url_param <> [] then raise Malformed_URL;
-	  if url'.url_query <> None then raise Malformed_URL;
+	  (* if url'.url_query <> None then raise Malformed_URL; 
+	   * Allowed since RFC 2396
+	   *)
 	end
     | ["";""] ->
 	(* This is illegal. *)
@@ -591,6 +742,7 @@ let null_url =
     url_validity = true;
     url_scheme   = None;
     url_user     = None;
+    url_user_param = [];
     url_password = None;
     url_host     = None;
     url_port     = None;
@@ -607,6 +759,7 @@ let make_url
       ?(encoded = false)
       ?scheme
       ?user
+      ?user_param
       ?password
       ?host
       ?port
@@ -615,25 +768,26 @@ let make_url
       ?query
       ?fragment
       ?other
-      url_syntax
+      syntax
   =
 
-  if not (url_syntax_is_valid url_syntax) then
+  if not (url_syntax_is_valid syntax) then
     invalid_arg "Neturl.make_url";
 
   modify_url
-    ~encoded:encoded
-    ~syntax:url_syntax
-    ?scheme:scheme
-    ?user:user
-    ?password:password
-    ?host:host
-    ?port:port
-    ?path:path
-    ?param:param
-    ?query:query
-    ?fragment:fragment
-    ?other:other
+    ~encoded
+    ~syntax
+    ?scheme
+    ?user
+    ?user_param
+    ?password
+    ?host
+    ?port
+    ?path
+    ?param
+    ?query
+    ?fragment
+    ?other
     null_url
 ;;
 
@@ -641,6 +795,7 @@ let make_url
 let remove_from_url
       ?(scheme = false)
       ?(user = false)
+      ?(user_param = false)
       ?(password = false)
       ?(host = false)
       ?(port = false)
@@ -653,17 +808,18 @@ let remove_from_url
   =
 
   make_url
-    ~encoded:  true
-    ?scheme:   (if scheme   then None else url.url_scheme)
-    ?user:     (if user     then None else url.url_user)
-    ?password: (if password then None else url.url_password)
-    ?host:     (if host     then None else url.url_host)
-    ?port:     (if port     then None else url.url_port)
-    ?path:     (if path     then None else Some url.url_path)
-    ?param:    (if param    then None else Some url.url_param)
-    ?query:    (if query    then None else url.url_query)
-    ?fragment: (if fragment then None else url.url_fragment)
-    ?other:    (if other    then None else url.url_other)
+    ~encoded:   true
+    ?scheme:    (if scheme     then None else url.url_scheme)
+    ?user:      (if user       then None else url.url_user)
+    ?user_param:(if user_param then None else Some url.url_user_param)
+    ?password:  (if password   then None else url.url_password)
+    ?host:      (if host       then None else url.url_host)
+    ?port:      (if port       then None else url.url_port)
+    ?path:      (if path       then None else Some url.url_path)
+    ?param:     (if param      then None else Some url.url_param)
+    ?query:     (if query      then None else url.url_query)
+    ?fragment:  (if fragment   then None else url.url_fragment)
+    ?other:     (if other      then None else url.url_other)
     url.url_syntax
 ;;
 
@@ -672,6 +828,7 @@ let default_url
       ?(encoded = false)
       ?scheme
       ?user
+      ?(user_param = [])
       ?password
       ?host
       ?port
@@ -708,17 +865,18 @@ let default_url
   in
 
   make_url
-    ~encoded:  true
-    ?scheme:   (pass_if_missing url.url_scheme   scheme)
-    ?user:     (pass_if_missing url.url_user     (enc user))
-    ?password: (pass_if_missing url.url_password (enc password))
-    ?host:     (pass_if_missing url.url_host     host)
-    ?port:     (pass_if_missing url.url_port     port)
-    ~path:     (if url.url_path  = [] then enc_list path  else url.url_path)
-    ~param:    (if url.url_param = [] then enc_list param else url.url_param)
-    ?query:    (pass_if_missing url.url_query    (enc query))
-    ?fragment: (pass_if_missing url.url_fragment (enc fragment))
-    ?other:    (pass_if_missing url.url_other    (enc other))
+    ~encoded:   true
+    ?scheme:    (pass_if_missing url.url_scheme   scheme)
+    ?user:      (pass_if_missing url.url_user     (enc user))
+    ~user_param:(if url.url_user_param = [] then enc_list user_param else url.url_user_param)
+    ?password:  (pass_if_missing url.url_password (enc password))
+    ?host:      (pass_if_missing url.url_host     host)
+    ?port:      (pass_if_missing url.url_port     port)
+    ~path:      (if url.url_path  = [] then enc_list path  else url.url_path)
+    ~param:     (if url.url_param = [] then enc_list param else url.url_param)
+    ?query:     (pass_if_missing url.url_query    (enc query))
+    ?fragment:  (pass_if_missing url.url_fragment (enc fragment))
+    ?other:     (pass_if_missing url.url_other    (enc other))
     url.url_syntax
 ;;
 
@@ -726,6 +884,7 @@ let default_url
 let undefault_url
       ?scheme
       ?user
+      ?user_param
       ?password
       ?host
       ?port
@@ -750,30 +909,26 @@ let undefault_url
 		   current)
   in
 
+  let remove_list_if_matching current arg =
+    match arg with
+	None -> current
+      | Some x ->
+	  if x = current then [] else current
+  in
+
   make_url
-    ~encoded:  true
-    ?scheme:   (remove_if_matching url.url_scheme   scheme)
-    ?user:     (remove_if_matching url.url_user     user)
-    ?password: (remove_if_matching url.url_password password)
-    ?host:     (remove_if_matching url.url_host     host)
-    ?port:     (remove_if_matching url.url_port     port)
-    ~path:     (match path with
-		     None -> url.url_path
-		   | Some x ->
-		       if x = url.url_path then
-			 []
-		       else
-			 url.url_path)
-    ~param:    (match param with
-		     None -> url.url_param
-		   | Some x ->
-		       if x = url.url_param then
-			 []
-		       else
-			 url.url_param)
-    ?query:    (remove_if_matching url.url_query    query)
-    ?fragment: (remove_if_matching url.url_fragment fragment)
-    ?other:    (remove_if_matching url.url_other    other)
+    ~encoded:   true
+    ?scheme:    (remove_if_matching url.url_scheme   scheme)
+    ?user:      (remove_if_matching url.url_user     user)
+    ~user_param:(remove_list_if_matching url.url_user_param user_param)
+    ?password:  (remove_if_matching url.url_password password)
+    ?host:      (remove_if_matching url.url_host     host)
+    ?port:      (remove_if_matching url.url_port     port)
+    ~path:      (remove_list_if_matching url.url_path path)
+    ~param:     (remove_list_if_matching url.url_param param)
+    ?query:     (remove_if_matching url.url_query    query)
+    ?fragment:  (remove_if_matching url.url_fragment fragment)
+    ?other:     (remove_if_matching url.url_other    other)
     url.url_syntax
 ;;
 
@@ -781,6 +936,7 @@ let undefault_url
 let url_provides 
       ?(scheme = false)
       ?(user = false)
+      ?(user_param = false)
       ?(password = false)
       ?(host = false)
       ?(port = false)
@@ -792,16 +948,17 @@ let url_provides
       url
   =
   
-  (scheme   => (url.url_scheme   <> None)) &
-  (user     => (url.url_user     <> None)) &
-  (password => (url.url_password <> None)) &
-  (host     => (url.url_host     <> None)) &
-  (port     => (url.url_port     <> None)) &
-  (path     => (url.url_path     <> []))   &
-  (param    => (url.url_param    <> [])) &
-  (query    => (url.url_query    <> None)) &
-  (fragment => (url.url_fragment <> None)) &
-  (other    => (url.url_other    <> None))
+  (scheme     => (url.url_scheme   <> None)) &&
+  (user       => (url.url_user     <> None)) &&
+  (user_param => (url.url_param    <> [])) &&
+  (password   => (url.url_password <> None)) &&
+  (host       => (url.url_host     <> None)) &&
+  (port       => (url.url_port     <> None)) &&
+  (path       => (url.url_path     <> [])) &&
+  (param      => (url.url_param    <> [])) &&
+  (query      => (url.url_query    <> None)) &&
+  (fragment   => (url.url_fragment <> None)) &&
+  (other      => (url.url_other    <> None))
 ;;
   
 
@@ -817,7 +974,7 @@ let decode_if want_encoded value =
   if want_encoded then
     value'
   else
-    Netencoding.Url.decode value'     (* WARNING: not thread-safe! *)
+    Netencoding.Url.decode value'
 ;;
 
 
@@ -826,20 +983,20 @@ let decode_path_if want_encoded value =
     value
   else
     List.map (Netencoding.Url.decode ~plus:true) value 
-        (* WARNING: not thread-safe! *)
 ;;
 
 
-let url_scheme                    url = return_if url.url_scheme;;
-let url_user     ?(encoded=false) url = decode_if encoded url.url_user;;
-let url_password ?(encoded=false) url = decode_if encoded url.url_password;;
-let url_host                      url = return_if url.url_host;;
-let url_port                      url = return_if url.url_port;;
-let url_path     ?(encoded=false) url = decode_path_if encoded url.url_path;;
-let url_param    ?(encoded=false) url = decode_path_if encoded url.url_param;;
-let url_query    ?(encoded=false) url = decode_if encoded url.url_query;;
-let url_fragment ?(encoded=false) url = decode_if encoded url.url_fragment;;
-let url_other    ?(encoded=false) url = decode_if encoded url.url_other;;
+let url_scheme                      url = return_if url.url_scheme;;
+let url_user       ?(encoded=false) url = decode_if encoded url.url_user;;
+let url_user_param ?(encoded=false) url = decode_path_if encoded url.url_user_param;;
+let url_password   ?(encoded=false) url = decode_if encoded url.url_password;;
+let url_host                        url = return_if url.url_host;;
+let url_port                        url = return_if url.url_port;;
+let url_path       ?(encoded=false) url = decode_path_if encoded url.url_path;;
+let url_param      ?(encoded=false) url = decode_path_if encoded url.url_param;;
+let url_query      ?(encoded=false) url = decode_if encoded url.url_query;;
+let url_fragment   ?(encoded=false) url = decode_if encoded url.url_fragment;;
+let url_other      ?(encoded=false) url = decode_if encoded url.url_other;;
 
 
 let string_of_url url =
@@ -856,6 +1013,10 @@ let string_of_url url =
 	      None -> "" 
 	    | Some user -> 
 		user ^ 
+		(String.concat ""
+		   (List.map
+		      (fun p -> ";" ^ p)
+		      url.url_user_param)) ^
 		(match url.url_password with
 		     None -> ""
 		   | Some password ->
@@ -899,6 +1060,9 @@ let string_of_url url =
 ;;
 
 
+let semi_re = Netstring_str.regexp ";";;
+
+
 let url_of_string url_syntax s =
   let l = String.length s in
   let recognized x = x <> Url_part_not_recognized in
@@ -937,11 +1101,11 @@ let url_of_string url_syntax s =
   in
 
   (* If there is a "//", a host will follow: *)
-  let host, port, user, password, k2 =
+  let host, port, userinfo, password, k2 =
     if recognized url_syntax.url_enable_host  &&
        k1 + 2 <= l  &&  s.[k1]='/'  && s.[k1+1]='/' then begin
 
-      let word_sep_list, k' = collect_words [ '/'; '#' ] '/' login_cats (k1+2) 
+      let word_sep_list, k' = collect_words [ '/'; '?'; '#' ] '/' login_cats (k1+2) 
       in
           (* or raise Malformed_URL *)
 
@@ -949,17 +1113,17 @@ let url_of_string url_syntax s =
 	try int_of_string x with _ -> raise Malformed_URL in
 
       match word_sep_list with
-	  [ host, ('/'|'#') ] ->
+	  [ host, ('/'|'?'|'#') ] ->
 	    Some host, None, None, None, k'
-	| [ host, ':'; port, ('/'|'#') ] ->
+	| [ host, ':'; port, ('/'|'?'|'#') ] ->
 	    Some host, Some (int port), None, None, k'
-	| [ user, '@'; host, ('/'|'#') ] ->
+	| [ user, '@'; host, ('/'|'?'|'#') ] ->
 	    Some host, None, Some user, None, k'
-	| [ user, '@'; host, ':'; port, ('/'|'#') ] ->
+	| [ user, '@'; host, ':'; port, ('/'|'?'|'#') ] ->
 	    Some host, Some (int port), Some user, None, k'
-	| [ user, ':'; password, '@'; host, ('/'|'#') ] ->
+	| [ user, ':'; password, '@'; host, ('/'|'?'|'#') ] ->
 	    Some host, None, Some user, Some password, k'
-	| [ user, ':'; password, '@'; host, ':'; port, ('/'|'#') ] ->
+	| [ user, ':'; password, '@'; host, ':'; port, ('/'|'?'|'#') ] ->
 	    Some host, Some (int port), Some user, Some password, k'
 	| _ ->
 	    raise Malformed_URL
@@ -968,6 +1132,20 @@ let url_of_string url_syntax s =
       None, None, None, None, k1
   in
 
+  (* Separate user from user_param: *)
+  let user, user_param =
+    match userinfo with
+	None -> (None, [])
+      | Some u ->
+	  if recognized url_syntax.url_enable_user_param then (
+	    let l = Netstring_str.split_delim semi_re u in
+	    match l with
+		[] -> (Some "", [])
+	      | user :: user_param -> (Some user, user_param)
+	  )
+	  else (Some u, [])
+  in
+
   let path, k3 =
     if recognized url_syntax.url_enable_path  &&
        k2 < l  (*  &&  s.[k2]='/'  *)
@@ -1091,20 +1269,61 @@ let url_of_string url_syntax s =
 
   make_url
     ~encoded:true
-    ?scheme:scheme
-    ?user:user
-    ?password:password
-    ?host:host
-    ?port:port
-    ~path:path
-    ~param:param
-    ?query:query
-    ?fragment:fragment
-    ?other:other
+    ?scheme
+    ?user
+    ~user_param
+    ?password
+    ?host
+    ?port
+    ~path
+    ~param
+    ?query
+    ?fragment
+    ?other
     url_syntax
 ;;
 
 
+let parse_url ?(schemes = common_url_syntax)
+              ?base_syntax
+	      ?(accept_8bits = false)
+	      ?(enable_fragment = false)
+	      s =
+  let scheme =
+    try
+      Some (extract_url_scheme s)
+    with
+	Malformed_URL -> None in
+  let syntax =
+    match scheme with
+	None ->
+	  ( match base_syntax with
+		None -> 
+		  raise Malformed_URL
+	      | Some syn -> 
+		  partial_url_syntax syn
+	  )
+      | Some sch ->
+	  try
+	    Hashtbl.find schemes sch
+	  with
+	      Not_found -> raise Malformed_URL
+  in
+  let syntax' =
+    if accept_8bits then
+      { syntax with url_accepts_8bits = true } 
+    else
+      syntax in
+  let syntax'' =
+    if enable_fragment && 
+       syntax.url_enable_fragment = Url_part_not_recognized then
+      { syntax' with url_enable_fragment = Url_part_allowed } 
+    else
+      syntax' in
+  url_of_string syntax'' s
+;;
+
+
 let split_path s =
   let l = String.length s in
   let rec collect_words k =
@@ -1214,6 +1433,7 @@ let apply_relative_url baseurl relurl =
 	      ?host:baseurl.url_host
 	      ?port:baseurl.url_port
 	      ?user:baseurl.url_user
+	      ~user_param:baseurl.url_user_param
 	      ?password:baseurl.url_password
 	      relurl
 	| [] ->
@@ -1241,6 +1461,7 @@ let apply_relative_url baseurl relurl =
 	      ?host:baseurl.url_host
 	      ?port:baseurl.url_port
 	      ?user:baseurl.url_user
+	      ~user_param:baseurl.url_user_param
 	      ?password:baseurl.url_password
 	      ~path:baseurl.url_path
 	      ~param:new_params
@@ -1268,6 +1489,7 @@ let apply_relative_url baseurl relurl =
 	      ?host:baseurl.url_host
 	      ?port:baseurl.url_port
 	      ?user:baseurl.url_user
+	      ~user_param:baseurl.url_user_param
 	      ?password:baseurl.url_password
 	      ~path:new_path               (* and change path *)
 	      relurl
@@ -1275,15 +1497,219 @@ let apply_relative_url baseurl relurl =
 ;;
 
 
+let ensure_absolute_url ?base u =
+  if u.url_scheme = None then (
+    match base with
+	None -> raise Malformed_URL
+      | Some b -> apply_relative_url b u
+  )
+  else
+    u
+;;
+  
+
 let print_url url =
   Format.print_string ("<URL:" ^ string_of_url url ^ ">")
 ;;
 
 
+let backslash_re = Netstring_pcre.regexp "\\\\";;
+let drive_letter_re = Netstring_pcre.regexp "^([A-Za-z]):/";;
+let drive_letter_re' = Netstring_pcre.regexp "^/([A-Za-z]):/";;
+let unc_path_re = Netstring_pcre.regexp "^//([^/]+)(/|$)";;
+
+
+let os_type = Sys.os_type;;
+
+let classify_path p =
+  match os_type with
+      "Unix" ->
+	if p <> "" && p.[0] = '/' then
+	  `Absolute_local p
+	else
+	  `Relative p
+    | "Win32" ->
+	let p' = Netstring_pcre.global_replace backslash_re "/" p in
+	( match Netstring_pcre.string_match drive_letter_re p' 0 with
+	      Some m ->
+		`Absolute_local ("/" ^ p')
+	    | None ->
+		( match Netstring_pcre.string_match unc_path_re p' 0 with
+		      Some m ->
+			let host = Netstring_pcre.matched_group m 1 p' in
+			let host_e = Netstring_pcre.group_end m 1 in
+			let path = String.sub p' host_e (String.length p' - host_e) in
+			let path = if path = "" then "/" else path in
+			`Absolute_remote(host,path)
+		    | None ->
+			if p' <> "" && p'.[0] = '/' then
+			  `Relative_drive p'
+			else
+			  `Relative p'
+		)
+	)
+    | "Cygwin" ->
+	let p' = Netstring_pcre.global_replace backslash_re "/" p in
+	( match Netstring_pcre.string_match drive_letter_re p' 0 with
+	      Some m ->
+		let letter = Netstring_pcre.matched_group m 1 p' in
+		let rest = String.sub p' 2 (String.length p' - 2) in
+		`Absolute_local("/cygdrive/" ^ letter ^ rest)
+	    | None ->
+		( match Netstring_pcre.string_match unc_path_re p' 0 with
+		      Some m ->
+			let host = Netstring_pcre.matched_group m 1 p' in
+			let host_e = Netstring_pcre.group_end m 1 in
+			let path = String.sub p' host_e (String.length p' - host_e) in
+			let path = if path = "" then "/" else path in
+			`Absolute_remote(host,path)
+		    | None ->
+			if p' <> "" && p'.[0] = '/' then
+			  `Absolute_local p'
+			else
+			  `Relative p'
+		)
+	)
+    | _ ->
+	assert false
+;;
+
+let file_url_of_local_path ?(getcwd = Sys.getcwd) p =
+  (* Classify p, and make it absolute: *)
+  let p_class = classify_path p in
+  let p_abs_class =
+    match p_class with
+	`Relative r ->
+	  ( match classify_path (getcwd()) with
+		`Absolute_local l ->
+		  if l = "/" then 
+		    `Absolute_local("/" ^ r)
+		  else
+		    `Absolute_local(l ^ "/" ^ r)
+	      | `Absolute_remote(h,l) ->
+		  if l = "/" then 
+		    `Absolute_remote(h,"/" ^ r)
+		  else
+		    `Absolute_remote(h,l ^ "/" ^ r)
+	      | _ ->
+		  failwith "Neturl.file_url_of_local_path: cwd is not absolute"
+	  )
+      | `Relative_drive r ->
+	  ( match classify_path (getcwd()) with
+		`Absolute_local l ->
+		  ( match Netstring_pcre.string_match drive_letter_re' l 0 with
+			Some m ->
+			  let letter = Netstring_pcre.matched_group m 1 l in
+			  `Absolute_local("/" ^ letter ^ ":" ^ r)
+		      | None ->
+			  assert false
+		  )
+	      | `Absolute_remote(h,l) ->
+		  `Absolute_remote(h,r)
+	      | _ ->
+		  failwith "Neturl.file_url_of_local_path: cwd is not absolute"
+	  )
+      | other -> other
+  in
+  (* Generate the URL: *)
+  let syntax = { file_url_syntax with url_accepts_8bits = true } in
+  match p_abs_class with
+      `Absolute_local l ->
+	let path = split_path l in
+	make_url ~scheme:"file" ~host:"localhost" ~path syntax
+    | `Absolute_remote(host,l) ->
+	let path = split_path l in
+	make_url ~scheme:"file" ~host ~path syntax
+    | _ ->
+	assert false
+;;
+
+
+let drive_letter_comp_re = Netstring_pcre.regexp "^([A-Za-z])(:|\\|)";;
+
+let local_path_of_file_url u =
+  let local_path p =
+    if p = [] || List.hd p <> "" then
+      failwith "Neturl.local_path_of_file_url: URL is not absolute";
+    match os_type with
+	"Unix" ->
+	  join_path p
+      | "Win32" ->
+	  (* There must be a drive letter: *)
+	  ( match p with
+		("" :: drive :: rest) ->
+		  (match Netstring_pcre.string_match drive_letter_comp_re drive 0 with
+		       Some m ->
+			 let letter = Netstring_pcre.matched_group m 1 drive in
+			 let rest = if rest = [] then [""] else rest in
+			 join_path((letter ^ ":") :: rest)
+		     | None ->
+			 failwith "Neturl.local_path_of_file_url: URL is not absolute"; 
+		  )
+	      | _ ->
+		  failwith "Neturl.local_path_of_file_url: URL is not absolute"; 
+	  )
+      | "Cygwin" ->
+	  (* Recognize drive letters: *)
+	  ( match p with
+		("" :: drive :: rest) ->
+		  (match Netstring_pcre.string_match drive_letter_comp_re drive 0 with
+		       Some m ->
+			 let letter = Netstring_pcre.matched_group m 1 drive in
+			 join_path("" :: "cygdrive" :: letter :: rest)
+		     | None ->
+			 join_path p
+		  )
+	      | _ ->
+		  join_path p
+	  )
+      | _ ->
+	  assert false
+  in
+  let remote_path host p =
+    if p = [] || List.hd p <> "" then
+      failwith "Neturl.local_path_of_file_url: URL is not absolute";
+    match os_type with
+	"Unix" ->
+	  failwith "Neturl.local_path_of_file_url: Cannot process non-local file URLs"
+      | "Win32" | "Cygwin" ->
+	  join_path( "" :: "" :: host :: List.tl p)
+      | _ ->
+	  assert false
+  in
+  let opt f =
+    try Some(f u) with Not_found -> None
+  in
+  match (opt url_scheme), (opt url_host), (url_path u) with
+      (Some "file", (Some("localhost"|"")|None), []) ->
+	local_path [""]
+    | (Some "file", (Some("localhost"|"")|None), p) ->
+	local_path p
+    | (Some "file", Some host, []) ->
+	remote_path host [""]
+    | (Some "file", Some host, p) ->
+	remote_path host p
+    | (Some _, _, _) ->
+	failwith "Neturl.local_path_of_file_url: Unexpected scheme"
+    | (None, _, _) ->
+	failwith "Neturl.local_path_of_file_url: Missing scheme (relative URL?)"
+;;
+
+
 (* ======================================================================
  * History:
  * 
  * $Log: neturl.ml,v $
+ * Revision 2.3  2004/07/18 16:36:24  stolpmann
+ * 	New: Neturl.file_url_of_local_path, local_path_of_file_url.
+ * 	URL updates according to RFC 2396.
+ *
+ * Revision 2.2  2004/07/12 22:57:01  stolpmann
+ * 	Added: parse_url, ensure_absolute_url
+ * 	Support for userparam.
+ * 	Syntax can prevent relative URLs.
+ * 	A lot of new syntaxes (pop, imap, ...)
+ *
  * Revision 2.1  2001/09/14 14:22:34  stolpmann
  * 	Initial revision (sourceforge)
  *
